from enum import Enum, IntEnum, IntFlag, auto

kCFAllocatorDefault = 0
MACH_PORT_NULL = 0
XPC_ARRAY_APPEND = 0xFFFFFFFFFFFFFFFF

# Block ABI flags from Block_private.h
BLOCK_DEALLOCATING = 0x0001
BLOCK_REFCOUNT_MASK = 0xFFFE
BLOCK_NEEDS_FREE = 1 << 24
BLOCK_HAS_COPY_DISPOSE = 1 << 25
BLOCK_HAS_CTOR = 1 << 26
BLOCK_IS_GC = 1 << 27
BLOCK_IS_GLOBAL = 1 << 28
BLOCK_HAS_STRET = 1 << 29
BLOCK_HAS_SIGNATURE = 1 << 30
BLOCK_HAS_EXTENDED_LAYOUT = 1 << 31

TASK_FLAVOR_CONTROL = 0  # a task_t
TASK_FLAVOR_READ = 1  # a task_read_t
TASK_FLAVOR_INSPECT = 2  # a task_inspect_t
TASK_FLAVOR_NAME = 3  # a task_name_t

EXC_TYPES_COUNT = 14  # incl. illegal exception 0
EXC_MASK_MACHINE = 0
EXCEPTION_CODE_MAX = 2  # code and subcode
EXC_BAD_ACCESS = 1  # Could not access memory
EXC_BAD_INSTRUCTION = 2  # Instruction failed
EXC_ARITHMETIC = 3  # Arithmetic exception
EXC_EMULATION = 4  # Emulation instruction
EXC_SOFTWARE = 5  # Software generated exception
EXC_BREAKPOINT = 6  # Trace, breakpoint, etc.
EXC_SYSCALL = 7  # System calls.
EXC_MACH_SYSCALL = 8  # Mach system calls.
EXC_RPC_ALERT = 9  # RPC alert
EXC_CRASH = 10  # Abnormal process exit
EXC_RESOURCE = 11  # Hit resource consumption limit
EXC_GUARD = 12  # Violated guarded resource protections
EXC_CORPSE_NOTIFY = 13  # Abnormal process exited to corpse state
EXCEPTION_DEFAULT = 1
EXCEPTION_STATE = 2
EXCEPTION_STATE_IDENTITY = 3
EXCEPTION_IDENTITY_PROTECTED = 4
MACH_EXCEPTION_BACKTRACE_PREFERRED = 0x20000000
MACH_EXCEPTION_ERRORS = 0x40000000
MACH_EXCEPTION_CODES = 0x80000000
MACH_EXCEPTION_MASK = MACH_EXCEPTION_CODES | MACH_EXCEPTION_ERRORS | MACH_EXCEPTION_BACKTRACE_PREFERRED
EXC_MASK_BAD_ACCESS = 1 << EXC_BAD_ACCESS
EXC_MASK_BAD_INSTRUCTION = 1 << EXC_BAD_INSTRUCTION
EXC_MASK_ARITHMETIC = 1 << EXC_ARITHMETIC
EXC_MASK_EMULATION = 1 << EXC_EMULATION
EXC_MASK_SOFTWARE = 1 << EXC_SOFTWARE
EXC_MASK_BREAKPOINT = 1 << EXC_BREAKPOINT
EXC_MASK_SYSCALL = 1 << EXC_SYSCALL
EXC_MASK_MACH_SYSCALL = 1 << EXC_MACH_SYSCALL
EXC_MASK_RPC_ALERT = 1 << EXC_RPC_ALERT
EXC_MASK_CRASH = 1 << EXC_CRASH
EXC_MASK_RESOURCE = 1 << EXC_RESOURCE
EXC_MASK_GUARD = 1 << EXC_GUARD
EXC_MASK_CORPSE_NOTIFY = 1 << EXC_CORPSE_NOTIFY

EXC_MASK_ALL = (
    EXC_MASK_BAD_ACCESS
    | EXC_MASK_BAD_INSTRUCTION
    | EXC_MASK_ARITHMETIC
    | EXC_MASK_EMULATION
    | EXC_MASK_SOFTWARE
    | EXC_MASK_BREAKPOINT
    | EXC_MASK_SYSCALL
    | EXC_MASK_MACH_SYSCALL
    | EXC_MASK_RPC_ALERT
    | EXC_MASK_RESOURCE
    | EXC_MASK_GUARD
    | EXC_MASK_MACHINE
)

IKOT_NONE = 0
IKOT_THREAD_CONTROL = 1
IKOT_TASK_CONTROL = 2
IKOT_HOST = 3
IKOT_HOST_PRIV = 4
IKOT_PROCESSOR = 5
IKOT_PSET = 6
IKOT_PSET_NAME = 7
IKOT_TIMER = 8
IKOT_PAGING_REQUEST = 9
IKOT_MIG = 10
IKOT_MEMORY_OBJECT = 11
IKOT_XMM_PAGER = 12
IKOT_XMM_KERNEL = 13
IKOT_XMM_REPLY = 14
IKOT_UND_REPLY = 15
IKOT_HOST_NOTIFY = 16
IKOT_HOST_SECURITY = 17
IKOT_LEDGER = 18
IKOT_MAIN_DEVICE = 19
IKOT_TASK_NAME = 20
IKOT_SUBSYSTEM = 21
IKOT_IO_DONE_QUEUE = 22
IKOT_SEMAPHORE = 23
IKOT_LOCK_SET = 24
IKOT_CLOCK = 25
IKOT_CLOCK_CTRL = 26
IKOT_IOKIT_IDENT = 27
IKOT_NAMED_ENTRY = 28
IKOT_IOKIT_CONNECT = 29
IKOT_IOKIT_OBJECT = 30
IKOT_UPL = 31
IKOT_MEM_OBJ_CONTROL = 32
IKOT_AU_SESSIONPORT = 33
IKOT_FILEPORT = 34
IKOT_LABELH = 35
IKOT_TASK_RESUME = 36
IKOT_VOUCHER = 37
IKOT_VOUCHER_ATTR_CONTROL = 38
IKOT_WORK_INTERVAL = 39
IKOT_UX_HANDLER = 40
IKOT_UEXT_OBJECT = 41
IKOT_ARCADE_REG = 42
IKOT_EVENTLINK = 43
IKOT_TASK_INSPECT = 44
IKOT_TASK_READ = 45
IKOT_THREAD_INSPECT = 46
IKOT_THREAD_READ = 47
IKOT_SUID_CRED = 48
IKOT_HYPERVISOR = 49

IKOT_UNKNOWN = 50  # magic catchall
IKOT_MAX_TYPE = IKOT_UNKNOWN + 1  # of IKOT_ types

MACH_PORT_RIGHT_SEND = 0
MACH_PORT_RIGHT_RECEIVE = 1
MACH_PORT_RIGHT_SEND_ONCE = 2
MACH_PORT_RIGHT_PORT_SET = 3
MACH_PORT_RIGHT_DEAD_NAME = 4
MACH_PORT_RIGHT_LABELH = 5  # obsolete right
MACH_PORT_RIGHT_NUMBER = 6  # right not implemented


def MACH_PORT_TYPE(right: int) -> int:
    return 1 << (right + 16)


MACH_PORT_TYPE_NONE = 0
MACH_PORT_TYPE_SEND = MACH_PORT_TYPE(MACH_PORT_RIGHT_SEND)
MACH_PORT_TYPE_RECEIVE = MACH_PORT_TYPE(MACH_PORT_RIGHT_RECEIVE)
MACH_PORT_TYPE_SEND_ONCE = MACH_PORT_TYPE(MACH_PORT_RIGHT_SEND_ONCE)
MACH_PORT_TYPE_PORT_SET = MACH_PORT_TYPE(MACH_PORT_RIGHT_PORT_SET)
MACH_PORT_TYPE_DEAD_NAME = MACH_PORT_TYPE(MACH_PORT_RIGHT_DEAD_NAME)
MACH_PORT_TYPE_LABELH = MACH_PORT_TYPE(MACH_PORT_RIGHT_LABELH)  # obsolete

# Convenient combinations

MACH_PORT_TYPE_SEND_RECEIVE = MACH_PORT_TYPE_SEND | MACH_PORT_TYPE_RECEIVE
MACH_PORT_TYPE_SEND_RIGHTS = MACH_PORT_TYPE_SEND | MACH_PORT_TYPE_SEND_ONCE
MACH_PORT_TYPE_PORT_RIGHTS = MACH_PORT_TYPE_SEND_RIGHTS | MACH_PORT_TYPE_RECEIVE
MACH_PORT_TYPE_PORT_OR_DEAD = MACH_PORT_TYPE_PORT_RIGHTS | MACH_PORT_TYPE_DEAD_NAME
MACH_PORT_TYPE_ALL_RIGHTS = MACH_PORT_TYPE_PORT_OR_DEAD | MACH_PORT_TYPE_PORT_SET

MACH_PORT_TYPE_DNREQUEST = 0x80000000
MACH_PORT_TYPE_SPREQUEST = 0x40000000
MACH_PORT_TYPE_SPREQUEST_DELAYED = 0x20000000

THREAD_IDENTIFIER_INFO = 4


class AVAudioSessionRouteSharingPolicy(IntEnum):
    Default = 0
    LongFormAudio = 1
    Independent = 2
    LongFormVideo = 3
    LongForm = LongFormAudio


class AVAudioSessionCategoryOptions(IntFlag):
    MixWithOthers = 0x1
    DuckOthers = 0x2
    AllowBluetooth = 0x4
    DefaultToSpeaker = 0x8
    InterruptSpokenAudioAndMixWithOthers = 0x11
    AllowBluetoothA2DP = 0x20
    AllowAirPlay = 0x40
    OverrideMutedMicrophoneInterruption = 0x80


class CFPropertyListFormat(IntEnum):
    kCFPropertyListOpenStepFormat = 1
    kCFPropertyListXMLFormat_v1_0 = 100
    kCFPropertyListBinaryFormat_v1_0 = 200


class CFPropertyListMutabilityOptions(IntEnum):
    kCFPropertyListImmutable = 0
    kCFPropertyListMutableContainers = (1 << 0,)
    kCFPropertyListMutableContainersAndLeaves = (1 << 1,)


class IOPMUserActiveType(IntEnum):
    kIOPMUserActiveLocal = 0  # User is local on the system
    kIOPMUserActiveRemote = 1  # Remote User connected to the system


kIOPMAssertionLevelOff = 0
kIOPMAssertionLevelOn = 255

# Types from MacTypes.h

# Basic C types
kCFNumberSInt8Type = 1
kCFNumberSInt16Type = 2
kCFNumberSInt32Type = 3
kCFNumberSInt64Type = 4
kCFNumberFloat32Type = 5
kCFNumberFloat64Type = 6  # 64-bit IEEE 754

# Other
kCFNumberCharType = 7
kCFNumberShortType = 8
kCFNumberIntType = 9
kCFNumberLongType = 10
kCFNumberLongLongType = 11
kCFNumberFloatType = 12
kCFNumberDoubleType = 13
kCFNumberCFIndexType = 14
kCFNumberMaxType = 14

# Types from IOKitKeys.h

# registry plane names
kIOServicePlane = "IOService"
kIOPowerPlane = "IOPower"
kIODeviceTreePlane = "IODeviceTree"
kIOAudioPlane = "IOAudio"
kIOFireWirePlane = "IOFireWire"
kIOUSBPlane = "IOUSB"

VM_REGION_INFO_MAX = 1024
VM_REGION_BASIC_INFO_64 = 9
VM_REGION_BASIC_INFO = 10

VM_FLAGS_FIXED = 0x0000
VM_FLAGS_ANYWHERE = 0x0001
VM_FLAGS_PURGABLE = 0x0002
VM_FLAGS_NO_PMAP_CHECK = 0x0004
VM_FLAGS_OVERWRITE = 0x0008

TASK_DYLD_INFO = 17
TASK_VM_INFO = 22

THREAD_BASIC_INFO = 3

# the i386_xxxx form is kept for legacy purposes since these types
# are externally known... eventually they should be deprecated.
# our internal implementation has moved to the following naming convention
#
#   x86_xxxx32 names are used to deal with 32 bit states
#   x86_xxxx64 names are used to deal with 64 bit states
#   x86_xxxx   names are used to deal with either 32 or 64 bit states via a self-describing mechanism
#

# these are the legacy names which should be deprecated in the future
# they are externally known which is the only reason we don't just get
# rid of them
#
i386_THREAD_STATE = 1
i386_FLOAT_STATE = 2
i386_EXCEPTION_STATE = 3

# THREAD_STATE_FLAVOR_LIST 0
# 	these are the supported flavors
#
x86_THREAD_STATE32 = 1
x86_FLOAT_STATE32 = 2
x86_EXCEPTION_STATE32 = 3
x86_THREAD_STATE64 = 4
x86_FLOAT_STATE64 = 5
x86_EXCEPTION_STATE64 = 6
x86_THREAD_STATE = 7
x86_FLOAT_STATE = 8
x86_EXCEPTION_STATE = 9
x86_DEBUG_STATE32 = 10
x86_DEBUG_STATE64 = 11
x86_DEBUG_STATE = 12
THREAD_STATE_NONE = 13


class ARMThreadFlavors(Enum):
    ARM_THREAD_STATE = 1
    ARM_VFP_STATE = 2
    ARM_EXCEPTION_STATE = 3
    ARM_DEBUG_STATE = 4
    ARN_THREAD_STATE_NONE = 5
    ARM_THREAD_STATE64 = 6
    ARM_EXCEPTION_STATE64 = 7


class CFStringEncoding(Enum):
    kCFStringEncodingMacRoman = 0
    kCFStringEncodingWindowsLatin1 = 0x0500  # ANSI codepage 1252
    kCFStringEncodingISOLatin1 = 0x0201  # ISO 8859-1
    kCFStringEncodingNextStepLatin = 0x0B01  # NextStep encoding
    kCFStringEncodingASCII = 0x0600  # 0..127 (in creating CFString values greater than 0x7F are

    # kTextEncodingUnicodeDefault  + kTextEncodingDefaultFormat (aka kUnicode16BitFormat)
    kCFStringEncodingUnicode = 0x0100

    kCFStringEncodingUTF8 = 0x08000100  # kTextEncodingUnicodeDefault + kUnicodeUTF8Format
    kCFStringEncodingNonLossyASCII = 0x0BFF  # 7bit Unicode variants used by Cocoa & Java

    # kTextEncodingUnicodeDefault + kUnicodeUTF16Format (alias of kCFStringEncodingUnicode)
    kCFStringEncodingUTF16 = 0x0100

    kCFStringEncodingUTF16BE = 0x10000100  # kTextEncodingUnicodeDefault + kUnicodeUTF16BEFormat
    kCFStringEncodingUTF16LE = 0x14000100  # kTextEncodingUnicodeDefault + kUnicodeUTF16LEFormat

    kCFStringEncodingUTF32 = 0x0C000100  # kTextEncodingUnicodeDefault + kUnicodeUTF32Format
    kCFStringEncodingUTF32BE = 0x18000100  # kTextEncodingUnicodeDefault + kUnicodeUTF32BEFormat
    kCFStringEncodingUTF32LE = 0x1C000100  # kTextEncodingUnicodeDefault + kUnicodeUTF32LEFormat


class NSStringEncoding(Enum):
    NSASCIIStringEncoding = 1  # 0..127 only
    NSNEXTSTEPStringEncoding = 2
    NSJapaneseEUCStringEncoding = 3
    NSUTF8StringEncoding = 4
    NSISOLatin1StringEncoding = 5
    NSSymbolStringEncoding = 6
    NSNonLossyASCIIStringEncoding = 7
    NSShiftJISStringEncoding = 8  # kCFStringEncodingDOSJapanese
    NSISOLatin2StringEncoding = 9
    NSUnicodeStringEncoding = 10
    NSWindowsCP1251StringEncoding = 11  # Cyrillic; same as AdobeStandardCyrillic
    NSWindowsCP1252StringEncoding = 12  # WinLatin1
    NSWindowsCP1253StringEncoding = 13  # Greek
    NSWindowsCP1254StringEncoding = 14  # Turkish
    NSWindowsCP1250StringEncoding = 15  # WinLatin2
    NSISO2022JPStringEncoding = 21  # ISO 2022 Japanese encoding for e-mail
    NSMacOSRomanStringEncoding = 30

    NSUTF16StringEncoding = NSUnicodeStringEncoding  # An alias for NSUnicodeStringEncoding

    NSUTF16BigEndianStringEncoding = 0x90000100  # NSUTF16StringEncoding encoding with explicit endianness specified
    NSUTF16LittleEndianStringEncoding = 0x94000100  # NSUTF16StringEncoding encoding with explicit endianness specified

    NSUTF32StringEncoding = 0x8C000100
    NSUTF32BigEndianStringEncoding = 0x98000100  # NSUTF32StringEncoding encoding with explicit endianness specified
    NSUTF32LittleEndianStringEncoding = 0x9C000100  # NSUTF32StringEncoding encoding with explicit endianness specified


# Taken from:
# https://opensource.apple.com/source/IOHIDFamily/IOHIDFamily-421.6/IOHIDFamily/IOHIDUsageTables.h.auto.html

kHIDPage_Undefined = 0x00
kHIDPage_GenericDesktop = 0x01
kHIDPage_Simulation = 0x02
kHIDPage_VR = 0x03
kHIDPage_Sport = 0x04
kHIDPage_Game = 0x05
# Reserved 0x06

# USB Device Class Definition for Human Interface Devices (HID). Note: the usage type for all key codes
# is Selector (Sel).
kHIDPage_KeyboardOrKeypad = 0x07
kHIDPage_LEDs = 0x08
kHIDPage_Button = 0x09
kHIDPage_Ordinal = 0x0A
kHIDPage_Telephony = 0x0B
kHIDPage_Consumer = 0x0C
kHIDPage_Digitizer = 0x0D
# Reserved 0x0E
kHIDPage_PID = 0x0F  # USB Physical Interface Device definitions for force feedback and related devices.
kHIDPage_Unicode = 0x10
# Reserved 0x11 - 0x13
kHIDPage_AlphanumericDisplay = 0x14
# Reserved 0x15 - 0x7F
# Monitor 0x80 - 0x83	 USB Device Class Definition for Monitor Devices
# Power 0x84 - 0x87	 USB Device Class Definition for Power Devices
kHIDPage_PowerDevice = 0x84  # Power Device Page
kHIDPage_BatterySystem = 0x85  # Battery System Page
# Reserved 0x88 - 0x8B
kHIDPage_BarCodeScanner = 0x8C  # (Point of Sale) USB Device Class Definition for Bar Code Scanner Devices
kHIDPage_WeighingDevice = 0x8D  # (Point of Sale) USB Device Class Definition for Weighing Devices
kHIDPage_Scale = 0x8D  # (Point of Sale) USB Device Class Definition for Scale Devices
kHIDPage_MagneticStripeReader = 0x8E
# ReservedPointofSalepages 0x8F
kHIDPage_CameraControl = 0x90  # USB Device Class Definition for Image Class Devices
kHIDPage_Arcade = 0x91  # OAAF Definitions for arcade and coinop related Devices
# Reserved 0x92 - 0xFEFF
# VendorDefined 0xFF00 - 0xFFFF
kHIDPage_VendorDefinedStart = 0xFF00

kHIDUsage_Undefined = 0x00

kHIDUsage_KeyboardErrorRollOver = 0x01  # ErrorRollOver
kHIDUsage_KeyboardPOSTFail = 0x02  # POSTFail
kHIDUsage_KeyboardErrorUndefined = 0x03  # ErrorUndefined
kHIDUsage_KeyboardA = 0x04  # a or A
kHIDUsage_KeyboardB = 0x05  # b or B
kHIDUsage_KeyboardC = 0x06  # c or C
kHIDUsage_KeyboardD = 0x07  # d or D
kHIDUsage_KeyboardE = 0x08  # e or E
kHIDUsage_KeyboardF = 0x09  # f or F
kHIDUsage_KeyboardG = 0x0A  # g or G
kHIDUsage_KeyboardH = 0x0B  # h or H
kHIDUsage_KeyboardI = 0x0C  # i or I
kHIDUsage_KeyboardJ = 0x0D  # j or J
kHIDUsage_KeyboardK = 0x0E  # k or K
kHIDUsage_KeyboardL = 0x0F  # l or L
kHIDUsage_KeyboardM = 0x10  # m or M
kHIDUsage_KeyboardN = 0x11  # n or N
kHIDUsage_KeyboardO = 0x12  # o or O
kHIDUsage_KeyboardP = 0x13  # p or P
kHIDUsage_KeyboardQ = 0x14  # q or Q
kHIDUsage_KeyboardR = 0x15  # r or R
kHIDUsage_KeyboardS = 0x16  # s or S
kHIDUsage_KeyboardT = 0x17  # t or T
kHIDUsage_KeyboardU = 0x18  # u or U
kHIDUsage_KeyboardV = 0x19  # v or V
kHIDUsage_KeyboardW = 0x1A  # w or W
kHIDUsage_KeyboardX = 0x1B  # x or X
kHIDUsage_KeyboardY = 0x1C  # y or Y
kHIDUsage_KeyboardZ = 0x1D  # z or Z
kHIDUsage_Keyboard1 = 0x1E  # 1 or !
kHIDUsage_Keyboard2 = 0x1F  # 2 or @
kHIDUsage_Keyboard3 = 0x20  # 3 or #
kHIDUsage_Keyboard4 = 0x21  # 4 or $
kHIDUsage_Keyboard5 = 0x22  # 5 or %
kHIDUsage_Keyboard6 = 0x23  # 6 or ^
kHIDUsage_Keyboard7 = 0x24  # 7 or &
kHIDUsage_Keyboard8 = 0x25  # 8 or *
kHIDUsage_Keyboard9 = 0x26  # 9 or (
kHIDUsage_Keyboard0 = 0x27  # 0 or )
kHIDUsage_KeyboardReturnOrEnter = 0x28  # Return (Enter)
kHIDUsage_KeyboardEscape = 0x29  # Escape
kHIDUsage_KeyboardDeleteOrBackspace = 0x2A  # Delete (Backspace)
kHIDUsage_KeyboardTab = 0x2B  # Tab
kHIDUsage_KeyboardSpacebar = 0x2C  # Spacebar
kHIDUsage_KeyboardHyphen = 0x2D  # - or _
kHIDUsage_KeyboardEqualSign = 0x2E  # = or +
kHIDUsage_KeyboardOpenBracket = 0x2F  # [ or {
kHIDUsage_KeyboardCloseBracket = 0x30  # ] or }
kHIDUsage_KeyboardBackslash = 0x31  # \ or |
kHIDUsage_KeyboardNonUSPound = 0x32  # Non-US # or _
kHIDUsage_KeyboardSemicolon = 0x33  # ; or :
kHIDUsage_KeyboardQuote = 0x34  # ' or "
kHIDUsage_KeyboardGraveAccentAndTilde = 0x35  # Grave Accent and Tilde
kHIDUsage_KeyboardComma = 0x36  # or <
kHIDUsage_KeyboardPeriod = 0x37  # . or >
kHIDUsage_KeyboardSlash = 0x38  # / or ?
kHIDUsage_KeyboardCapsLock = 0x39  # Caps Lock
kHIDUsage_KeyboardF1 = 0x3A  # F1
kHIDUsage_KeyboardF2 = 0x3B  # F2
kHIDUsage_KeyboardF3 = 0x3C  # F3
kHIDUsage_KeyboardF4 = 0x3D  # F4
kHIDUsage_KeyboardF5 = 0x3E  # F5
kHIDUsage_KeyboardF6 = 0x3F  # F6
kHIDUsage_KeyboardF7 = 0x40  # F7
kHIDUsage_KeyboardF8 = 0x41  # F8
kHIDUsage_KeyboardF9 = 0x42  # F9
kHIDUsage_KeyboardF10 = 0x43  # F10
kHIDUsage_KeyboardF11 = 0x44  # F11
kHIDUsage_KeyboardF12 = 0x45  # F12
kHIDUsage_KeyboardPrintScreen = 0x46  # Print Screen
kHIDUsage_KeyboardScrollLock = 0x47  # Scroll Lock
kHIDUsage_KeyboardPause = 0x48  # Pause
kHIDUsage_KeyboardInsert = 0x49  # Insert
kHIDUsage_KeyboardHome = 0x4A  # Home
kHIDUsage_KeyboardPageUp = 0x4B  # Page Up
kHIDUsage_KeyboardDeleteForward = 0x4C  # Delete Forward
kHIDUsage_KeyboardEnd = 0x4D  # End
kHIDUsage_KeyboardPageDown = 0x4E  # Page Down
kHIDUsage_KeyboardRightArrow = 0x4F  # Right Arrow
kHIDUsage_KeyboardLeftArrow = 0x50  # Left Arrow
kHIDUsage_KeyboardDownArrow = 0x51  # Down Arrow
kHIDUsage_KeyboardUpArrow = 0x52  # Up Arrow
kHIDUsage_KeypadNumLock = 0x53  # Keypad NumLock or Clear
kHIDUsage_KeypadSlash = 0x54  # Keypad /
kHIDUsage_KeypadAsterisk = 0x55  # Keypad *
kHIDUsage_KeypadHyphen = 0x56  # Keypad -
kHIDUsage_KeypadPlus = 0x57  # Keypad +
kHIDUsage_KeypadEnter = 0x58  # Keypad Enter
kHIDUsage_Keypad1 = 0x59  # Keypad 1 or End
kHIDUsage_Keypad2 = 0x5A  # Keypad 2 or Down Arrow
kHIDUsage_Keypad3 = 0x5B  # Keypad 3 or Page Down
kHIDUsage_Keypad4 = 0x5C  # Keypad 4 or Left Arrow
kHIDUsage_Keypad5 = 0x5D  # Keypad 5
kHIDUsage_Keypad6 = 0x5E  # Keypad 6 or Right Arrow
kHIDUsage_Keypad7 = 0x5F  # Keypad 7 or Home
kHIDUsage_Keypad8 = 0x60  # Keypad 8 or Up Arrow
kHIDUsage_Keypad9 = 0x61  # Keypad 9 or Page Up
kHIDUsage_Keypad0 = 0x62  # Keypad 0 or Insert
kHIDUsage_KeypadPeriod = 0x63  # Keypad . or Delete
kHIDUsage_KeyboardNonUSBackslash = 0x64  # Non-US \ or |
kHIDUsage_KeyboardApplication = 0x65  # Application
kHIDUsage_KeyboardPower = 0x66  # Power
kHIDUsage_KeypadEqualSign = 0x67  # Keypad =
kHIDUsage_KeyboardF13 = 0x68  # F13
kHIDUsage_KeyboardF14 = 0x69  # F14
kHIDUsage_KeyboardF15 = 0x6A  # F15
kHIDUsage_KeyboardF16 = 0x6B  # F16
kHIDUsage_KeyboardF17 = 0x6C  # F17
kHIDUsage_KeyboardF18 = 0x6D  # F18
kHIDUsage_KeyboardF19 = 0x6E  # F19
kHIDUsage_KeyboardF20 = 0x6F  # F20
kHIDUsage_KeyboardF21 = 0x70  # F21
kHIDUsage_KeyboardF22 = 0x71  # F22
kHIDUsage_KeyboardF23 = 0x72  # F23
kHIDUsage_KeyboardF24 = 0x73  # F24
kHIDUsage_KeyboardExecute = 0x74  # Execute
kHIDUsage_KeyboardHelp = 0x75  # Help
kHIDUsage_KeyboardMenu = 0x76  # Menu
kHIDUsage_KeyboardSelect = 0x77  # Select
kHIDUsage_KeyboardStop = 0x78  # Stop
kHIDUsage_KeyboardAgain = 0x79  # Again
kHIDUsage_KeyboardUndo = 0x7A  # Undo
kHIDUsage_KeyboardCut = 0x7B  # Cut
kHIDUsage_KeyboardCopy = 0x7C  # Copy
kHIDUsage_KeyboardPaste = 0x7D  # Paste
kHIDUsage_KeyboardFind = 0x7E  # Find
kHIDUsage_KeyboardMute = 0x7F  # Mute
kHIDUsage_KeyboardVolumeUp = 0x80  # Volume Up
kHIDUsage_KeyboardVolumeDown = 0x81  # Volume Down
kHIDUsage_KeyboardLockingCapsLock = 0x82  # Locking Caps Lock
kHIDUsage_KeyboardLockingNumLock = 0x83  # Locking Num Lock
kHIDUsage_KeyboardLockingScrollLock = 0x84  # Locking Scroll Lock
kHIDUsage_KeypadComma = 0x85  # Keypad Comma
kHIDUsage_KeypadEqualSignAS400 = 0x86  # Keypad Equal Sign for AS/400
kHIDUsage_KeyboardInternational1 = 0x87  # International1
kHIDUsage_KeyboardInternational2 = 0x88  # International2
kHIDUsage_KeyboardInternational3 = 0x89  # International3
kHIDUsage_KeyboardInternational4 = 0x8A  # International4
kHIDUsage_KeyboardInternational5 = 0x8B  # International5
kHIDUsage_KeyboardInternational6 = 0x8C  # International6
kHIDUsage_KeyboardInternational7 = 0x8D  # International7
kHIDUsage_KeyboardInternational8 = 0x8E  # International8
kHIDUsage_KeyboardInternational9 = 0x8F  # International9
kHIDUsage_KeyboardLANG1 = 0x90  # LANG1
kHIDUsage_KeyboardLANG2 = 0x91  # LANG2
kHIDUsage_KeyboardLANG3 = 0x92  # LANG3
kHIDUsage_KeyboardLANG4 = 0x93  # LANG4
kHIDUsage_KeyboardLANG5 = 0x94  # LANG5
kHIDUsage_KeyboardLANG6 = 0x95  # LANG6
kHIDUsage_KeyboardLANG7 = 0x96  # LANG7
kHIDUsage_KeyboardLANG8 = 0x97  # LANG8
kHIDUsage_KeyboardLANG9 = 0x98  # LANG9
kHIDUsage_KeyboardAlternateErase = 0x99  # AlternateErase
kHIDUsage_KeyboardSysReqOrAttention = 0x9A  # SysReq/Attention
kHIDUsage_KeyboardCancel = 0x9B  # Cancel
kHIDUsage_KeyboardClear = 0x9C  # Clear
kHIDUsage_KeyboardPrior = 0x9D  # Prior
kHIDUsage_KeyboardReturn = 0x9E  # Return
kHIDUsage_KeyboardSeparator = 0x9F  # Separator
kHIDUsage_KeyboardOut = 0xA0  # Out
kHIDUsage_KeyboardOper = 0xA1  # Oper
kHIDUsage_KeyboardClearOrAgain = 0xA2  # Clear/Again
kHIDUsage_KeyboardCrSelOrProps = 0xA3  # CrSel/Props
kHIDUsage_KeyboardExSel = 0xA4  # ExSel
# 0xA5-0xDF Reserved
kHIDUsage_KeyboardLeftControl = 0xE0  # Left Control
kHIDUsage_KeyboardLeftShift = 0xE1  # Left Shift
kHIDUsage_KeyboardLeftAlt = 0xE2  # Left Alt
kHIDUsage_KeyboardLeftGUI = 0xE3  # Left GUI
kHIDUsage_KeyboardRightControl = 0xE4  # Right Control
kHIDUsage_KeyboardRightShift = 0xE5  # Right Shift
kHIDUsage_KeyboardRightAlt = 0xE6  # Right Alt
kHIDUsage_KeyboardRightGUI = 0xE7  # Right GUI
# 0xE8-0xFFFF Reserved
kHIDUsage_Keyboard_Reserved = 0xFFFF

kHIDUsage_Csmr_ConsumerControl = 0x01  # Application Collection
kHIDUsage_Csmr_NumericKeyPad = 0x02  # Named Array
kHIDUsage_Csmr_ProgrammableButtons = 0x03  # Named Array
# 0x03 - 0x1F Reserved
kHIDUsage_Csmr_Plus10 = 0x20  # One-Shot Control
kHIDUsage_Csmr_Plus100 = 0x21  # One-Shot Control
kHIDUsage_Csmr_AMOrPM = 0x22  # One-Shot Control
# 0x23 - 0x3F Reserved
kHIDUsage_Csmr_Power = 0x30  # On/Off Control
kHIDUsage_Csmr_Reset = 0x31  # One-Shot Control
kHIDUsage_Csmr_Sleep = 0x32  # One-Shot Control
kHIDUsage_Csmr_SleepAfter = 0x33  # One-Shot Control
kHIDUsage_Csmr_SleepMode = 0x34  # Re-Trigger Control
kHIDUsage_Csmr_Illumination = 0x35  # On/Off Control
kHIDUsage_Csmr_FunctionButtons = 0x36  # Named Array
# 0x37 - 0x3F Reserved
kHIDUsage_Csmr_Menu = 0x40  # On/Off Control
kHIDUsage_Csmr_MenuPick = 0x41  # One-Shot Control
kHIDUsage_Csmr_MenuUp = 0x42  # One-Shot Control
kHIDUsage_Csmr_MenuDown = 0x43  # One-Shot Control
kHIDUsage_Csmr_MenuLeft = 0x44  # One-Shot Control
kHIDUsage_Csmr_MenuRight = 0x45  # One-Shot Control
kHIDUsage_Csmr_MenuEscape = 0x46  # One-Shot Control
kHIDUsage_Csmr_MenuValueIncrease = 0x47  # One-Shot Control
kHIDUsage_Csmr_MenuValueDecrease = 0x48  # One-Shot Control
# 0x49 - 0x5F Reserved
kHIDUsage_Csmr_DataOnScreen = 0x60  # On/Off Control
kHIDUsage_Csmr_ClosedCaption = 0x61  # On/Off Control
kHIDUsage_Csmr_ClosedCaptionSelect = 0x62  # Selector
kHIDUsage_Csmr_VCROrTV = 0x63  # On/Off Control
kHIDUsage_Csmr_BroadcastMode = 0x64  # One-Shot Control
kHIDUsage_Csmr_Snapshot = 0x65  # One-Shot Control
kHIDUsage_Csmr_Still = 0x66  # One-Shot Control
# 0x67 - 0x7F Reserved
kHIDUsage_Csmr_Selection = 0x80  # Named Array
kHIDUsage_Csmr_Assign = 0x81  # Selector
kHIDUsage_Csmr_ModeStep = 0x82  # One-Shot Control
kHIDUsage_Csmr_RecallLast = 0x83  # One-Shot Control
kHIDUsage_Csmr_EnterChannel = 0x84  # One-Shot Control
kHIDUsage_Csmr_OrderMovie = 0x85  # One-Shot Control
kHIDUsage_Csmr_Channel = 0x86  # Linear Control
kHIDUsage_Csmr_MediaSelection = 0x87  # Selector
kHIDUsage_Csmr_MediaSelectComputer = 0x88  # Selector
kHIDUsage_Csmr_MediaSelectTV = 0x89  # Selector
kHIDUsage_Csmr_MediaSelectWWW = 0x8A  # Selector
kHIDUsage_Csmr_MediaSelectDVD = 0x8B  # Selector
kHIDUsage_Csmr_MediaSelectTelephone = 0x8C  # Selector
kHIDUsage_Csmr_MediaSelectProgramGuide = 0x8D  # Selector
kHIDUsage_Csmr_MediaSelectVideoPhone = 0x8E  # Selector
kHIDUsage_Csmr_MediaSelectGames = 0x8F  # Selector
kHIDUsage_Csmr_MediaSelectMessages = 0x90  # Selector
kHIDUsage_Csmr_MediaSelectCD = 0x91  # Selector
kHIDUsage_Csmr_MediaSelectVCR = 0x92  # Selector
kHIDUsage_Csmr_MediaSelectTuner = 0x93  # Selector
kHIDUsage_Csmr_Quit = 0x94  # One-Shot Control
kHIDUsage_Csmr_Help = 0x95  # On/Off Control
kHIDUsage_Csmr_MediaSelectTape = 0x96  # Selector
kHIDUsage_Csmr_MediaSelectCable = 0x97  # Selector
kHIDUsage_Csmr_MediaSelectSatellite = 0x98  # Selector
kHIDUsage_Csmr_MediaSelectSecurity = 0x99  # Selector
kHIDUsage_Csmr_MediaSelectHome = 0x9A  # Selector
kHIDUsage_Csmr_MediaSelectCall = 0x9B  # Selector
kHIDUsage_Csmr_ChannelIncrement = 0x9C  # One-Shot Control
kHIDUsage_Csmr_ChannelDecrement = 0x9D  # One-Shot Control
kHIDUsage_Csmr_Media = 0x9E  # Selector
# 0x9F Reserved
kHIDUsage_Csmr_VCRPlus = 0xA0  # One-Shot Control
kHIDUsage_Csmr_Once = 0xA1  # One-Shot Control
kHIDUsage_Csmr_Daily = 0xA2  # One-Shot Control
kHIDUsage_Csmr_Weekly = 0xA3  # One-Shot Control
kHIDUsage_Csmr_Monthly = 0xA4  # One-Shot Control
# 0xA5 - 0xAF Reserved
kHIDUsage_Csmr_Play = 0xB0  # On/Off Control
kHIDUsage_Csmr_Pause = 0xB1  # On/Off Control
kHIDUsage_Csmr_Record = 0xB2  # On/Off Control
kHIDUsage_Csmr_FastForward = 0xB3  # On/Off Control
kHIDUsage_Csmr_Rewind = 0xB4  # On/Off Control
kHIDUsage_Csmr_ScanNextTrack = 0xB5  # One-Shot Control
kHIDUsage_Csmr_ScanPreviousTrack = 0xB6  # One-Shot Control
kHIDUsage_Csmr_Stop = 0xB7  # One-Shot Control
kHIDUsage_Csmr_Eject = 0xB8  # One-Shot Control
kHIDUsage_Csmr_RandomPlay = 0xB9  # On/Off Control
kHIDUsage_Csmr_SelectDisc = 0xBA  # Named Array
kHIDUsage_Csmr_EnterDisc = 0xBB  # Momentary Control
kHIDUsage_Csmr_Repeat = 0xBC  # One-Shot Control
kHIDUsage_Csmr_Tracking = 0xBD  # Linear Control
kHIDUsage_Csmr_TrackNormal = 0xBE  # One-Shot Control
kHIDUsage_Csmr_SlowTracking = 0xBF  # Linear Control
kHIDUsage_Csmr_FrameForward = 0xC0  # Re-Trigger Control
kHIDUsage_Csmr_FrameBack = 0xC1  # Re-Trigger Control
kHIDUsage_Csmr_Mark = 0xC2  # One-Shot Control
kHIDUsage_Csmr_ClearMark = 0xC3  # One-Shot Control
kHIDUsage_Csmr_RepeatFromMark = 0xC4  # On/Off Control
kHIDUsage_Csmr_ReturnToMark = 0xC5  # One-Shot Control
kHIDUsage_Csmr_SearchMarkForward = 0xC6  # One-Shot Control
kHIDUsage_Csmr_SearchMarkBackwards = 0xC7  # One-Shot Control
kHIDUsage_Csmr_CounterReset = 0xC8  # One-Shot Control
kHIDUsage_Csmr_ShowCounter = 0xC9  # One-Shot Control
kHIDUsage_Csmr_TrackingIncrement = 0xCA  # Re-Trigger Control
kHIDUsage_Csmr_TrackingDecrement = 0xCB  # Re-Trigger Control
kHIDUsage_Csmr_StopOrEject = 0xCC  # One-Shot Control
kHIDUsage_Csmr_PlayOrPause = 0xCD  # One-Shot Control
kHIDUsage_Csmr_PlayOrSkip = 0xCE  # One-Shot Control
# 0xCF - 0xDF Reserved
kHIDUsage_Csmr_Volume = 0xE0  # Linear Control
kHIDUsage_Csmr_Balance = 0xE1  # Linear Control
kHIDUsage_Csmr_Mute = 0xE2  # On/Off Control
kHIDUsage_Csmr_Bass = 0xE3  # Linear Control
kHIDUsage_Csmr_Treble = 0xE4  # Linear Control
kHIDUsage_Csmr_BassBoost = 0xE5  # On/Off Control
kHIDUsage_Csmr_SurroundMode = 0xE6  # One-Shot Control
kHIDUsage_Csmr_Loudness = 0xE7  # On/Off Control
kHIDUsage_Csmr_MPX = 0xE8  # On/Off Control
kHIDUsage_Csmr_VolumeIncrement = 0xE9  # Re-Trigger Control
kHIDUsage_Csmr_VolumeDecrement = 0xEA  # Re-Trigger Control
# 0xEB - 0xEF Reserved
kHIDUsage_Csmr_Speed = 0xF0  # Selector
kHIDUsage_Csmr_PlaybackSpeed = 0xF1  # Named Array
kHIDUsage_Csmr_StandardPlay = 0xF2  # Selector
kHIDUsage_Csmr_LongPlay = 0xF3  # Selector
kHIDUsage_Csmr_ExtendedPlay = 0xF4  # Selector
kHIDUsage_Csmr_Slow = 0xF5  # One-Shot Control
# 0xF6 - 0xFF Reserved
kHIDUsage_Csmr_FanEnable = 0x100  # On/Off Control
kHIDUsage_Csmr_FanSpeed = 0x101  # Linear Control
kHIDUsage_Csmr_LightEnable = 0x102  # On/Off Control
kHIDUsage_Csmr_LightIlluminationLevel = 0x103  # Linear Control
kHIDUsage_Csmr_ClimateControlEnable = 0x104  # On/Off Control
kHIDUsage_Csmr_RoomTemperature = 0x105  # Linear Control
kHIDUsage_Csmr_SecurityEnable = 0x106  # On/Off Control
kHIDUsage_Csmr_FireAlarm = 0x107  # One-Shot Control
kHIDUsage_Csmr_PoliceAlarm = 0x108  # One-Shot Control
# 0x109 - 0x14F Reserved
kHIDUsage_Csmr_BalanceRight = 0x150  # Re-Trigger Control
kHIDUsage_Csmr_BalanceLeft = 0x151  # Re-Trigger Control
kHIDUsage_Csmr_BassIncrement = 0x152  # Re-Trigger Control
kHIDUsage_Csmr_BassDecrement = 0x153  # Re-Trigger Control
kHIDUsage_Csmr_TrebleIncrement = 0x154  # Re-Trigger Control
kHIDUsage_Csmr_TrebleDecrement = 0x155  # Re-Trigger Control
# 0x156 - 0x15F Reserved
kHIDUsage_Csmr_SpeakerSystem = 0x160  # Logical Collection
kHIDUsage_Csmr_ChannelLeft = 0x161  # Logical Collection
kHIDUsage_Csmr_ChannelRight = 0x162  # Logical Collection
kHIDUsage_Csmr_ChannelCenter = 0x163  # Logical Collection
kHIDUsage_Csmr_ChannelFront = 0x164  # Logical Collection
kHIDUsage_Csmr_ChannelCenterFront = 0x165  # Logical Collection
kHIDUsage_Csmr_ChannelSide = 0x166  # Logical Collection
kHIDUsage_Csmr_ChannelSurround = 0x167  # Logical Collection
kHIDUsage_Csmr_ChannelLowFrequencyEnhancement = 0x168  # Logical Collection
kHIDUsage_Csmr_ChannelTop = 0x169  # Logical Collection
kHIDUsage_Csmr_ChannelUnknown = 0x16A  # Logical Collection
# 0x16B - 0x16F Reserved
kHIDUsage_Csmr_SubChannel = 0x170  # Linear Control
kHIDUsage_Csmr_SubChannelIncrement = 0x171  # One-Shot Control
kHIDUsage_Csmr_SubChannelDecrement = 0x172  # One-Shot Control
kHIDUsage_Csmr_AlternateAudioIncrement = 0x173  # One-Shot Control
kHIDUsage_Csmr_AlternateAudioDecrement = 0x174  # One-Shot Control
# 0x175 - 0x17F Reserved
kHIDUsage_Csmr_ApplicationLaunchButtons = 0x180  # Named Array
kHIDUsage_Csmr_ALLaunchButtonConfigurationTool = 0x181  # Selector
kHIDUsage_Csmr_ALProgrammableButtonConfiguration = 0x182  # Selector
kHIDUsage_Csmr_ALConsumerControlConfiguration = 0x183  # Selector
kHIDUsage_Csmr_ALWordProcessor = 0x184  # Selector
kHIDUsage_Csmr_ALTextEditor = 0x185  # Selector
kHIDUsage_Csmr_ALSpreadsheet = 0x186  # Selector
kHIDUsage_Csmr_ALGraphicsEditor = 0x187  # Selector
kHIDUsage_Csmr_ALPresentationApp = 0x188  # Selector
kHIDUsage_Csmr_ALDatabaseApp = 0x189  # Selector
kHIDUsage_Csmr_ALEmailReader = 0x18A  # Selector
kHIDUsage_Csmr_ALNewsreader = 0x18B  # Selector
kHIDUsage_Csmr_ALVoicemail = 0x18C  # Selector
kHIDUsage_Csmr_ALContactsOrAddressBook = 0x18D  # Selector
kHIDUsage_Csmr_ALCalendarOrSchedule = 0x18E  # Selector
kHIDUsage_Csmr_ALTaskOrProjectManager = 0x18F  # Selector
kHIDUsage_Csmr_ALLogOrJournalOrTimecard = 0x190  # Selector
kHIDUsage_Csmr_ALCheckbookOrFinance = 0x191  # Selector
kHIDUsage_Csmr_ALCalculator = 0x192  # Selector
kHIDUsage_Csmr_ALAOrVCaptureOrPlayback = 0x193  # Selector
kHIDUsage_Csmr_ALLocalMachineBrowser = 0x194  # Selector
kHIDUsage_Csmr_ALLANOrWANBrowser = 0x195  # Selector
kHIDUsage_Csmr_ALInternetBrowser = 0x196  # Selector
kHIDUsage_Csmr_ALRemoteNetworkingOrISPConnect = 0x197  # Selector
kHIDUsage_Csmr_ALNetworkConference = 0x198  # Selector
kHIDUsage_Csmr_ALNetworkChat = 0x199  # Selector
kHIDUsage_Csmr_ALTelephonyOrDialer = 0x19A  # Selector
kHIDUsage_Csmr_ALLogon = 0x19B  # Selector
kHIDUsage_Csmr_ALLogoff = 0x19C  # Selector
kHIDUsage_Csmr_ALLogonOrLogoff = 0x19D  # Selector
kHIDUsage_Csmr_ALTerminalLockOrScreensaver = 0x19E  # Selector
kHIDUsage_Csmr_ALControlPanel = 0x19F  # Selector
kHIDUsage_Csmr_ALCommandLineProcessorOrRun = 0x1A0  # Selector
kHIDUsage_Csmr_ALProcessOrTaskManager = 0x1A1  # Selector
kHIDUsage_Csmr_AL = 0x1A2  # Selector
kHIDUsage_Csmr_ALNextTaskOrApplication = 0x1A3  # Selector
kHIDUsage_Csmr_ALPreviousTaskOrApplication = 0x1A4  # Selector
kHIDUsage_Csmr_ALPreemptiveHaltTaskOrApplication = 0x1A5  # Selector
kHIDUsage_Csmr_ALIntegratedHelpCenter = 0x1A6  # Selector
kHIDUsage_Csmr_ALDocuments = 0x1A7  # Selector
kHIDUsage_Csmr_ALThesaurus = 0x1A8  # Selector
kHIDUsage_Csmr_ALDictionary = 0x1A9  # Selector
kHIDUsage_Csmr_ALDesktop = 0x1AA  # Selector
kHIDUsage_Csmr_ALSpellCheck = 0x1AB  # Selector
kHIDUsage_Csmr_ALGrammerCheck = 0x1AC  # Selector
kHIDUsage_Csmr_ALWirelessStatus = 0x1AD  # Selector
kHIDUsage_Csmr_ALKeyboardLayout = 0x1AE  # Selector
kHIDUsage_Csmr_ALVirusProtection = 0x1AF  # Selector
kHIDUsage_Csmr_ALEncryption = 0x1B0  # Selector
kHIDUsage_Csmr_ALScreenSaver = 0x1B1  # Selector
kHIDUsage_Csmr_ALAlarms = 0x1B2  # Selector
kHIDUsage_Csmr_ALClock = 0x1B3  # Selector
kHIDUsage_Csmr_ALFileBrowser = 0x1B4  # Selector
kHIDUsage_Csmr_ALPowerStatus = 0x1B5  # Selector
# 0x1A6 - 0x1FF Reserved
kHIDUsage_Csmr_GenericGUIApplicationControls = 0x200  # Named Array
kHIDUsage_Csmr_ACNew = 0x201  # Selector
kHIDUsage_Csmr_ACOpen = 0x202  # Selector
kHIDUsage_Csmr_ACClose = 0x203  # Selector
kHIDUsage_Csmr_ACExit = 0x204  # Selector
kHIDUsage_Csmr_ACMaximize = 0x205  # Selector
kHIDUsage_Csmr_ACMinimize = 0x206  # Selector
kHIDUsage_Csmr_ACSave = 0x207  # Selector
kHIDUsage_Csmr_ACPrint = 0x208  # Selector
kHIDUsage_Csmr_ACProperties = 0x209  # Selector
kHIDUsage_Csmr_ACUndo = 0x21A  # Selector
kHIDUsage_Csmr_ACCopy = 0x21B  # Selector
kHIDUsage_Csmr_ACCut = 0x21C  # Selector
kHIDUsage_Csmr_ACPaste = 0x21D  # Selector
kHIDUsage_Csmr_AC = 0x21E  # Selector
kHIDUsage_Csmr_ACFind = 0x21F  # Selector
kHIDUsage_Csmr_ACFindandReplace = 0x220  # Selector
kHIDUsage_Csmr_ACSearch = 0x221  # Selector
kHIDUsage_Csmr_ACGoTo = 0x222  # Selector
kHIDUsage_Csmr_ACHome = 0x223  # Selector
kHIDUsage_Csmr_ACBack = 0x224  # Selector
kHIDUsage_Csmr_ACForward = 0x225  # Selector
kHIDUsage_Csmr_ACStop = 0x226  # Selector
kHIDUsage_Csmr_ACRefresh = 0x227  # Selector
kHIDUsage_Csmr_ACPreviousLink = 0x228  # Selector
kHIDUsage_Csmr_ACNextLink = 0x229  # Selector
kHIDUsage_Csmr_ACBookmarks = 0x22A  # Selector
kHIDUsage_Csmr_ACHistory = 0x22B  # Selector
kHIDUsage_Csmr_ACSubscriptions = 0x22C  # Selector
kHIDUsage_Csmr_ACZoomIn = 0x22D  # Selector
kHIDUsage_Csmr_ACZoomOut = 0x22E  # Selector
kHIDUsage_Csmr_ACZoom = 0x22F  # Selector
kHIDUsage_Csmr_ACFullScreenView = 0x230  # Selector
kHIDUsage_Csmr_ACNormalView = 0x231  # Selector
kHIDUsage_Csmr_ACViewToggle = 0x232  # Selector
kHIDUsage_Csmr_ACScrollUp = 0x233  # Selector
kHIDUsage_Csmr_ACScrollDown = 0x234  # Selector
kHIDUsage_Csmr_ACScroll = 0x235  # Selector
kHIDUsage_Csmr_ACPanLeft = 0x236  # Selector
kHIDUsage_Csmr_ACPanRight = 0x237  # Selector
kHIDUsage_Csmr_ACPan = 0x238  # Selector
kHIDUsage_Csmr_ACNewWindow = 0x239  # Selector
kHIDUsage_Csmr_ACTileHorizontally = 0x23A  # Selector
kHIDUsage_Csmr_ACTileVertically = 0x23B  # Selector
kHIDUsage_Csmr_ACFormat = 0x23C  # Selector
# 0x23D - 0xFFFF Reserved
kHIDUsage_Csmr_Reserved = 0xFFFF

kSecCodeMagicRequirement = 0xFADE0C00  # single requirement
kSecCodeMagicRequirementSet = 0xFADE0C01  # requirement set
kSecCodeMagicCodeDirectory = 0xFADE0C02  # CodeDirectory
kSecCodeMagicEmbeddedSignature = 0xFADE0CC0  # single-architecture embedded signature
kSecCodeMagicDetachedSignature = 0xFADE0CC1  # detached multi-architecture signature
kSecCodeMagicEntitlement = 0xFADE7171  # entitlement blob


class IOHIDDigitizerTransducerType(Enum):
    kIOHIDDigitizerTransducerTypeStylus = 0
    kIOHIDDigitizerTransducerTypePuck = 1
    kIOHIDDigitizerTransducerTypeFinger = 2
    kIOHIDDigitizerTransducerTypeHand = 3


class IOHIDSwipeMask(Enum):
    kIOHIDSwipeNone = 0x00000000
    kIOHIDSwipeUp = 0x00000001
    kIOHIDSwipeDown = 0x00000002
    kIOHIDSwipeLeft = 0x00000004
    kIOHIDSwipeRight = 0x00000008
    kIOHIDScaleExpand = 0x00000010
    kIOHIDScaleContract = 0x00000020
    kIOHIDRotateCW = 0x00000040
    kIOHIDRotateCCW = 0x00000080


def IOHIDEventTypeMask(type_):
    return 1 << type_


def IOHIDEventFieldBase(type_):
    return type_ << 16


def IOHIDEventFieldOffsetOf(field):
    return field & 0xFFFF


class IOHIDEventType(Enum):
    kIOHIDEventTypeNULL = 0
    kIOHIDEventTypeVendorDefined = 1
    kIOHIDEventTypeButton = 2
    kIOHIDEventTypeKeyboard = 3
    kIOHIDEventTypeTranslation = 4
    kIOHIDEventTypeRotation = 5
    kIOHIDEventTypeScroll = 6
    kIOHIDEventTypeScale = 7
    kIOHIDEventTypeZoom = 8
    kIOHIDEventTypeVelocity = 9
    kIOHIDEventTypeOrientation = 10
    kIOHIDEventTypeDigitizer = 11
    kIOHIDEventTypeAmbientLightSensor = 12
    kIOHIDEventTypeAccelerometer = 13
    kIOHIDEventTypeProximity = 14
    kIOHIDEventTypeTemperature = 15
    kIOHIDEventTypeNavigationSwipe = 16
    kIOHIDEventTypePointer = 17
    kIOHIDEventTypeProgress = 18
    kIOHIDEventTypeMultiAxisPointer = 19
    kIOHIDEventTypeGyro = 20
    kIOHIDEventTypeCompass = 21
    kIOHIDEventTypeZoomToggle = 22
    kIOHIDEventTypeDockSwipe = 23  # just like kIOHIDEventTypeNavigationSwipe, but intended for consumption by Dock
    kIOHIDEventTypeSymbolicHotKey = 24
    kIOHIDEventTypePower = 25
    kIOHIDEventTypeLED = 26
    kIOHIDEventTypeFluidTouchGesture = 27  # This will eventually superseed Navagation and Dock swipes
    kIOHIDEventTypeBoundaryScroll = 28
    kIOHIDEventTypeBiometric = 29
    kIOHIDEventTypeUnicode = 30
    kIOHIDEventTypeAtmosphericPressure = 31
    kIOHIDEventTypeForce = 32
    kIOHIDEventTypeMotionActivity = 33
    kIOHIDEventTypeMotionGesture = 34
    kIOHIDEventTypeGameController = 35
    kIOHIDEventTypeHumidity = 36
    kIOHIDEventTypeCollection = 37
    kIOHIDEventTypeBrightness = 38
    kIOHIDEventTypeCount = 39  # This should always be last

    # DEPRECATED:
    kIOHIDEventTypeSwipe = 16
    kIOHIDEventTypeMouse = 17


class IOHIDDigitizerEventMask(Enum):
    kIOHIDDigitizerEventRange = 1 << 0
    kIOHIDDigitizerEventTouch = 1 << 1
    kIOHIDDigitizerEventPosition = 1 << 2
    kIOHIDDigitizerEventStop = 1 << 3
    kIOHIDDigitizerEventPeak = 1 << 4
    kIOHIDDigitizerEventIdentity = 1 << 5
    kIOHIDDigitizerEventAttribute = 1 << 6
    kIOHIDDigitizerEventCancel = 1 << 7
    kIOHIDDigitizerEventStart = 1 << 8
    kIOHIDDigitizerEventResting = 1 << 9
    kIOHIDDigitizerEventFromEdgeFlat = 1 << 10
    kIOHIDDigitizerEventFromEdgeTip = 1 << 11
    kIOHIDDigitizerEventFromCorner = 1 << 12
    kIOHIDDigitizerEventSwipePending = 1 << 13
    kIOHIDDigitizerEventFromEdgeForcePending = 1 << 14
    kIOHIDDigitizerEventFromEdgeForceActive = 1 << 15
    kIOHIDDigitizerEventForcePopped = 1 << 16
    kIOHIDDigitizerEventSwipeUp = 1 << 24
    kIOHIDDigitizerEventSwipeDown = 1 << 25
    kIOHIDDigitizerEventSwipeLeft = 1 << 26
    kIOHIDDigitizerEventSwipeRight = 1 << 27
    kIOHIDDigitizerEventEstimatedAltitude = 1 << 28
    kIOHIDDigitizerEventEstimatedAzimuth = 1 << 29
    kIOHIDDigitizerEventEstimatedPressure = 1 << 30
    kIOHIDDigitizerEventSwipeMask = 0xF << 24


class IOHIDEventField(Enum):
    kIOHIDEventFieldIsRelative = IOHIDEventFieldBase(IOHIDEventType.kIOHIDEventTypeNULL.value)
    kIOHIDEventFieldIsCollection = auto()
    kIOHIDEventFieldIsPixelUnits = auto()
    kIOHIDEventFieldIsCenterOrigin = auto()
    kIOHIDEventFieldIsBuiltIn = auto()


class IOHIDEventOptionBits(Enum):
    kIOHIDEventOptionNone = 0
    kIOHIDEventOptionIsAbsolute = 1 << 0
    kIOHIDEventOptionIsCollection = 1 << 1
    kIOHIDEventOptionIsPixelUnits = 1 << 2
    kIOHIDEventOptionIsCenterOrigin = 1 << 3
    kIOHIDEventOptionIsBuiltIn = 1 << 4

    # misspellings
    kIOHIDEventOptionPixelUnits = kIOHIDEventOptionIsPixelUnits


class IOHIDEventFieldDigitizer(Enum):
    kIOHIDEventFieldDigitizerX = IOHIDEventFieldBase(IOHIDEventType.kIOHIDEventTypeDigitizer.value)
    kIOHIDEventFieldDigitizerY = auto()
    kIOHIDEventFieldDigitizerZ = auto()
    kIOHIDEventFieldDigitizerButtonMask = auto()
    kIOHIDEventFieldDigitizerType = auto()
    kIOHIDEventFieldDigitizerIndex = auto()
    kIOHIDEventFieldDigitizerIdentity = auto()
    kIOHIDEventFieldDigitizerEventMask = auto()
    kIOHIDEventFieldDigitizerRange = auto()
    kIOHIDEventFieldDigitizerTouch = auto()
    kIOHIDEventFieldDigitizerPressure = auto()
    kIOHIDEventFieldDigitizerAuxiliaryPressure = auto()  # BarrelPressure
    kIOHIDEventFieldDigitizerTwist = auto()
    kIOHIDEventFieldDigitizerTiltX = auto()
    kIOHIDEventFieldDigitizerTiltY = auto()
    kIOHIDEventFieldDigitizerAltitude = auto()
    kIOHIDEventFieldDigitizerAzimuth = auto()
    kIOHIDEventFieldDigitizerQuality = auto()
    kIOHIDEventFieldDigitizerDensity = auto()
    kIOHIDEventFieldDigitizerIrregularity = auto()
    kIOHIDEventFieldDigitizerMajorRadius = auto()
    kIOHIDEventFieldDigitizerMinorRadius = auto()
    kIOHIDEventFieldDigitizerCollection = auto()
    kIOHIDEventFieldDigitizerCollectionChord = auto()
    kIOHIDEventFieldDigitizerChildEventMask = auto()
    kIOHIDEventFieldDigitizerIsDisplayIntegrated = auto()
    kIOHIDEventFieldDigitizerQualityRadiiAccuracy = auto()
    kIOHIDEventFieldDigitizerGenerationCount = auto()
    kIOHIDEventFieldDigitizerWillUpdateMask = auto()
    kIOHIDEventFieldDigitizerDidUpdateMask = auto()
    kIOHIDEventFieldDigitizerEstimatedMask = auto()


kCGHIDEventTap = 0
kCGWindowListOptionAll = 0
kCGNullWindowID = 0

CG_KEY_A = 0
CG_KEY_S = 1
CG_KEY_D = 2
CG_KEY_F = 3
CG_KEY_H = 4
CG_KEY_G = 5
CG_KEY_Z = 6
CG_KEY_X = 7
CG_KEY_C = 8
CG_KEY_V = 9
CG_KEY_B = 11
CG_KEY_Q = 12
CG_KEY_W = 13
CG_KEY_E = 14
CG_KEY_R = 15
CG_KEY_Y = 16
CG_KEY_T = 17
CG_KEY_1 = 18
CG_KEY_2 = 19
CG_KEY_3 = 20
CG_KEY_4 = 21
CG_KEY_6 = 22
CG_KEY_5 = 23
CG_KEY_EQUAL = 24
CG_KEY_9 = 25
CG_KEY_7 = 26
CG_KEY_DASH = 27
CG_KEY_8 = 28
CG_KEY_0 = 29
CG_KEY_CLOSE_SQUARE_BRACKETS = 30
CG_KEY_O = 31
CG_KEY_U = 32
CG_KEY_OPEN_SQUARE_BRACKETS = 33
CG_KEY_I = 34
CG_KEY_P = 35
CG_KEY_RETURN = 36
CG_KEY_L = 37
CG_KEY_J = 38
CG_KEY_COLON = 39
CG_KEY_K = 40
CG_KEY_SEMICOLON = 41
CG_KEY_BACKSLASH = 42
CG_KEY_COMMA = 43
CG_KEY_SLASH = 44
CG_KEY_N = 45
CG_KEY_M = 46
CG_KEY_DOT = 47
CG_KEY_TAB = 48
CG_KEY_SPACE = 49
CG_KEY_TILDE = 50
CG_KEY_DELETE = 51
CG_KEY_ENTER = 52
CG_KEY_ESCAPE = 53
CG_KEY_ASTERISK = 67
CG_KEY_PLUS = 69
CG_KEY_CLEAR = 71
CG_KEY_F5 = 96
CG_KEY_F6 = 97
CG_KEY_F7 = 98
CG_KEY_F3 = 99
CG_KEY_F8 = 100
CG_KEY_F9 = 101
CG_KEY_F11 = 103
CG_KEY_F13 = 105
CG_KEY_F14 = 107
CG_KEY_F10 = 109
CG_KEY_F12 = 111
CG_KEY_F15 = 113
CG_KEY_HELP = 114
CG_KEY_HOME = 115
CG_KEY_PGUP = 116
CG_KEY_F4 = 118
CG_KEY_END = 119
CG_KEY_F2 = 120
CG_KEY_PGDN = 121
CG_KEY_F1 = 122
CG_KEY_LEFT = 123
CG_KEY_RIGHT = 124
CG_KEY_DOWN = 125
CG_KEY_UP = 126


class OsLogLevel(IntEnum):
    """LogLevel. Extracted by reversing LevelForKey(id prefs, NSString *key)"""

    OFF = 0
    NONE = 0
    UNDEFINED = 1  # no such plist key
    DEFAULT = 2
    INFO = 3
    DEBUG = 4


# Definitions of flags stored in file flags word.
# Super-user and owner changeable flags.

UF_SETTABLE = 0x0000FFFF  # mask of owner changeable flags
UF_NODUMP = 0x00000001  # do not dump file
UF_IMMUTABLE = 0x00000002  # file may not be changed
UF_APPEND = 0x00000004  # writes to file may only append
UF_OPAQUE = 0x00000008  # directory is opaque wrt. union

# The following bit is reserved for FreeBSD.  It is not implemented
# in Mac OS X.

# UF_NOUNLINK =  0x00000010  # file may not be removed or renamed
UF_COMPRESSED = 0x00000020  # file is compressed (some file-systems)

# UF_TRACKED is used for dealing with document IDs.  We no longer issue
#  notifications for deletes or renames for files which have UF_TRACKED set.
UF_TRACKED = 0x00000040

UF_DATAVAULT = 0x00000080  # entitlement required for reading
# and writing

# Bits 0x0100 through 0x4000 are currently undefined.
UF_HIDDEN = 0x00008000  # hint that this item should not be
# displayed in a GUI
#
# Super-user changeable flags.

SF_SUPPORTED = 0x009F0000  # mask of superuser supported flags
SF_SETTABLE = 0x3FFF0000  # mask of superuser changeable flags
SF_SYNTHETIC = 0xC0000000  # mask of system read-only synthetic flags
SF_ARCHIVED = 0x00010000  # file is archived
SF_IMMUTABLE = 0x00020000  # file may not be changed
SF_APPEND = 0x00040000  # writes to file may only append
SF_RESTRICTED = 0x00080000  # entitlement required for writing
SF_NOUNLINK = 0x00100000  # Item may not be removed, renamed or mounted on

# The following two bits are reserved for FreeBSD.  They are not
# implemented in Mac OS X.

# SF_SNAPSHOT =  0x00200000  # snapshot inode
# NOTE: There is no SF_HIDDEN bit.

SF_FIRMLINK = 0x00800000  # file is a firmlink

# Synthetic flags.
#
# These are read-only.  We keep them out of SF_SUPPORTED so that
# attempts to set them will fail.

SF_DATALESS = 0x40000000  # file is dataless object

# CoreSymbolication.h
kCSNow = 0x8000000000000000



import logging
from typing import TYPE_CHECKING, Optional

from rpcclient.clients.darwin.consts import IOPMUserActiveType
from rpcclient.exceptions import BadReturnValueError

if TYPE_CHECKING:
    from rpcclient.clients.darwin.client import DarwinClient

logger = logging.getLogger(__name__)


class PowerAssertion:
    def __init__(self, client: "DarwinClient", identifier: int):
        self._client = client
        self.id = identifier

    def __enter__(self) -> "PowerAssertion":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self._client.symbols.IOPMAssertionRelease(self.id)


class Power:
    """Power utils"""

    def __init__(self, client: "DarwinClient"):
        self._client = client

    def declare_user_activity(self, name: str, type_: IOPMUserActiveType) -> PowerAssertion:
        """Declares that the user is active on the system"""
        with self._client.safe_malloc(8) as p_assertion_id:
            err = self._client.symbols.IOPMAssertionDeclareUserActivity(self._client.cf(name), type_, p_assertion_id)
            if err != 0:
                raise BadReturnValueError("IOPMAssertionCreateWithProperties() failed")
            return PowerAssertion(self._client, p_assertion_id[0])

    def declare_network_client_activity(self, name: str) -> PowerAssertion:
        """
        A convenience function for handling remote network clients; this is a wrapper for holding
        kIOPMAssertNetworkClientActive
        """
        with self._client.safe_malloc(8) as p_assertion_id:
            err = self._client.symbols.IOPMDeclareNetworkClientActivity(self._client.cf(name), p_assertion_id)
            if err != 0:
                raise BadReturnValueError("IOPMAssertionCreateWithProperties() failed")
            return PowerAssertion(self._client, p_assertion_id[0])

    def create_assertion(self, name: str, type_: str, reason: Optional[str] = None) -> PowerAssertion:
        properties = {"AssertName": name, "AssertType": type_}
        if reason is not None:
            properties["HumanReadableReason"] = reason
        with self._client.safe_malloc(8) as p_assertion_id:
            err = self._client.symbols.IOPMAssertionCreateWithProperties(self._client.cf(properties), p_assertion_id)
            if err != 0:
                raise BadReturnValueError("IOPMAssertionCreateWithProperties() failed")
            return PowerAssertion(self._client, p_assertion_id[0])

    def copy_assertions_by_process(self) -> dict[int, dict]:
        """Returns a dictionary listing all assertions, grouped by their owning process"""
        with self._client.safe_malloc(8) as p_assertions:
            if self._client.symbols.IOPMCopyAssertionsByProcess(p_assertions) != 0:
                raise BadReturnValueError("IOPMCopyAssertionsByProcess() failed")
            assertions = p_assertions[0]
        if not assertions:
            return {}
        result = {}
        key_enumerator = assertions.objc_call("keyEnumerator")
        while True:
            pid_object = key_enumerator.objc_call("nextObject")
            if pid_object == 0:
                break
            pid = pid_object.py()
            result[pid] = assertions.objc_call("objectForKey:", pid_object).py()
        return result

    def copy_scheduled_power_events(self) -> list[dict]:
        """List all scheduled system power events"""
        return self._client.symbols.IOPMCopyScheduledPowerEvents().py()

    def copy_assertions_status(self) -> dict[str, int]:
        """Returns a list of available assertions and their system-wide levels"""
        with self._client.safe_malloc(8) as result:
            if self._client.symbols.IOPMCopyAssertionsStatus(result) != 0:
                raise BadReturnValueError("IOPMCopyAssertionsStatus() failed")
            return result[0].py()

    def reboot(self) -> None:
        if self._client.symbols.reboot().c_int64 == -1:
            raise BadReturnValueError()

    def sleep(self) -> None:
        """
        Enter sustem sleep

        See: https://gist.github.com/atr000/416796
        """
        with self._client.safe_malloc(8) as p_master:
            err = self._client.symbols.IOMasterPort(self._client.symbols.bootstrap_port[0], p_master)
            if err != 0:
                raise BadReturnValueError(f"IOMasterPort didnt work, err is {err}")
            pmcon = self._client.symbols.IOPMFindPowerManagement(p_master[0])
            if pmcon == 0:
                raise BadReturnValueError("IOPMFindPowerManagement coudlnt establish connection")
            err = self._client.symbols.IOPMSleepSystem(pmcon)
            if err != 0:
                raise BadReturnValueError(f"IOPMSleepSystem didnt work. err is {err}")



import logging
from typing import TYPE_CHECKING

from rpcclient.exceptions import RpcSetDeveloperModeError

if TYPE_CHECKING:
    from rpcclient.clients.ios.client import IosClient

logger = logging.getLogger(__name__)


class Amfi:
    """AMFI utilities (AppleMobileFileIntegrity)."""

    def __init__(self, client: "IosClient") -> None:
        self._client = client
        self._client.load_framework("AppleMobileFileIntegrity")

    def set_developer_mode_status(self, enabled: bool) -> None:
        """Enable or disable Developer Mode via the AMFI XPC service."""
        cfreply = self._client.xpc.send_message_using_cf_serialization(
            "com.apple.amfi.xpc", {"action": int(not (enabled))}, False
        )["cfreply"]
        raw_response = next(iter(cfreply.values()))
        if b"success" not in raw_response:
            raise RpcSetDeveloperModeError()

    @property
    def developer_mode_status(self) -> bool:
        """Return True if Developer Mode is enabled."""
        return self._client.symbols.amfi_developer_mode_status()



from collections import namedtuple
from typing import TYPE_CHECKING

from rpcclient.clients.darwin.symbol import DarwinSymbol

if TYPE_CHECKING:
    from rpcclient.clients.ios.client import IosClient

CGRect = namedtuple("CGRect", "x0 y0 x1 y1")


class ScreenCapture:
    """Screen capture utilities."""

    def __init__(self, client: "IosClient") -> None:
        self._client = client

    @property
    def main_display(self) -> DarwinSymbol:
        """Return the main CADisplay instance."""
        return self._client.symbols.objc_getClass("CADisplay").objc_call("mainDisplay")

    @property
    def bounds(self) -> CGRect:
        """Return the main display bounds."""
        result = self.main_display.objc_call("bounds", return_raw=True)
        return CGRect(x0=result.d0, y0=result.d1, x1=result.d2, y1=result.d3)

    @property
    def screenshot(self) -> bytes:
        """Return a PNG screenshot of the current screen."""
        return self._client.symbols.UIImagePNGRepresentation(self._client.symbols._UICreateScreenUIImage()).py()



from datetime import datetime
from typing import TYPE_CHECKING

from rpcclient.clients.darwin.structs import timeval

if TYPE_CHECKING:
    from rpcclient.clients.darwin.client import DarwinClient


class Time:
    def __init__(self, client: "DarwinClient"):
        """
        :param rpcclient.darwin.client.DarwinClient client:
        """
        self._client = client
        self._client.load_framework("CoreTime")

    def now(self) -> datetime:
        """get current time"""
        with self._client.safe_calloc(timeval.sizeof()) as current:
            self._client.symbols.gettimeofday(current, 0)
            time_of_day = timeval.parse(current.peek(timeval.sizeof()))
        return datetime.fromtimestamp(time_of_day.tv_sec + (time_of_day.tv_usec / (10**6)))

    def set_current(self, new_time: datetime):
        """set current time"""
        self._client.symbols.TMSetAutomaticTimeZoneEnabled(0)
        with self._client.safe_calloc(timeval.sizeof()) as current:
            current.poke(timeval.build({"tv_sec": int(new_time.timestamp()), "tv_usec": new_time.microsecond}))
            self._client.symbols.settimeofday(current, 0)

    def set_auto(self):
        """opt-in automatic time settings"""
        self._client.symbols.TMSetAutomaticTimeZoneEnabled(1)

    @property
    def is_set_automatically(self):
        """tell is time settings are set to automatic"""
        return bool(self._client.symbols.TMIsAutomaticTimeZoneEnabled())

    def boot_time(self) -> datetime:
        timestamp = timeval.parse(self._client.sysctl.get_by_name("kern.boottime")).tv_sec
        return datetime.fromtimestamp(timestamp)



class RpcClientException(Exception):
    """general exception"""

    pass


class InvalidServerVersionMagicError(RpcClientException):
    """server handshake failed due to an invalid magic"""

    pass


class ResponseNotFoundError(RpcClientException):
    """Response not found"""

    pass


class SubsystemInitError(RpcClientException):
    """Raised when unavailable subsystems are used."""

    pass


class ServerResponseError(RpcClientException):
    """Server returned error"""

    pass


class ServerDiedError(RpcClientException):
    """server became disconnected during an operation"""

    pass


class SymbolAbsentError(RpcClientException):
    """trying to access a symbol which is not exported from any library currently loaded into the server's memory"""

    pass


class ProcessSymbolAbsentError(RpcClientException):
    """trying to access a symbol which is not exported from any library currently loaded into the process's memory"""

    pass


class ArgumentError(RpcClientException):
    """at least one of the supplied arguments for a given function was invalid"""

    pass


class BadReturnValueError(RpcClientException):
    """remote c function returned an error"""

    pass


class NoSuchPreferenceError(RpcClientException):
    """attempt to read a preference data which doesn't exist"""

    pass


class CfSerializationError(RpcClientException):
    """failed to encode/decode a cfobject into/from a python object"""

    pass


class SpawnError(RpcClientException):
    """failed to spawn a child process"""

    pass


class FailedToConnectError(RpcClientException):
    """failed to connect to rpcserver"""

    pass


class UnrecognizedSelectorError(RpcClientException):
    """tried to access a non-existing objc object selector"""

    pass


class GettingObjectiveCClassError(RpcClientException):
    """failed to create an objc class wrapper for a given object"""

    pass


class MissingLibraryError(RpcClientException):
    """a required library could not be found"""

    pass


class NoEntitlementsError(RpcClientException):
    """binary contains no entitlements"""

    pass


class HarGlobalNotFoundError(RpcClientException):
    """Failed to find Harlogger global"""

    pass


class ElementNotFoundError(RpcClientException):
    """Failed to find element"""

    pass


class FirstElementNotFoundError(ElementNotFoundError):
    """Failed to find the first element"""

    pass


class LastElementNotFoundError(ElementNotFoundError):
    """Failed to find the last element"""

    pass


class LaunchError(BadReturnValueError):
    """failed to launch a child process"""

    pass


class RpcFileExistsError(BadReturnValueError):
    """RPC version for FileExistsError (errno = EEXIST)"""

    pass


class RpcFileNotFoundError(BadReturnValueError):
    """RPC version for FileNotFoundError (errno = ENOENTRY)"""

    pass


class RpcBrokenPipeError(BadReturnValueError):
    """RPC version for BrokenPipeError (errno = EPIPE)"""

    pass


class RpcNotEmptyError(BadReturnValueError):
    """raised when errno = ENOTEMPTY"""

    pass


class RpcIsADirectoryError(BadReturnValueError):
    """RPC version for IsADirectoryError (errno = ENOTEMPTY)"""

    pass


class RpcNotADirectoryError(BadReturnValueError):
    """RPC version for NotADirectoryError (errno = ENOTDIR)"""

    pass


class RpcPermissionError(BadReturnValueError):
    """RPC version for PermissionError (errno = EPERM)"""

    pass


class RpcAccessibilityTurnedOffError(BadReturnValueError):
    """Application AX and Automation is turned off"""

    pass


class RpcFailedToRecordError(BadReturnValueError):
    """An attempt to record has failed"""

    pass


class RpcFailedToPlayError(BadReturnValueError):
    """An attempt to play has failed"""

    pass


class RpcResourceTemporarilyUnavailableError(BadReturnValueError):
    """raised when errno = EAGAIN"""

    pass


class RpcConnectionRefusedError(BadReturnValueError):
    """RPC version for ConnectionRefusedError (errno = ECONNREFUSED)"""

    pass


class RpcDeadClientError(BadReturnValueError):
    """RPC client in a dead state"""

    pass


class NoSuchClientError(RpcClientException):
    pass


class NoSuchActivityError(RpcClientException):
    pass


class RpcFailedLaunchingAppError(BadReturnValueError):
    """Failed to launch application"""

    pass


class RpcAppleScriptError(BadReturnValueError):
    """Failed to execute given AppleScript"""

    pass


class RpcXpcError(BadReturnValueError):
    """XPC-related error"""

    pass


class RpcXpcSerializationError(RpcXpcError):
    """Failed to serialize/deserialize XPC message"""

    pass


class RpcSetDeveloperModeError(BadReturnValueError):
    """Failed to set Developer Mode"""

    pass


class RpcFailedToGetPrimaryAppError(BadReturnValueError):
    """Failed to get [AXElement primaryApp]"""

    pass



import struct
from enum import Enum
from typing import TYPE_CHECKING

from parameter_decorators import path_to_str

from rpcclient.clients.darwin.consts import AVAudioSessionCategoryOptions, AVAudioSessionRouteSharingPolicy
from rpcclient.clients.darwin.symbol import DarwinSymbol
from rpcclient.core.allocated import Allocated
from rpcclient.exceptions import BadReturnValueError, RpcFailedToPlayError, RpcFailedToRecordError

if TYPE_CHECKING:
    from rpcclient.clients.darwin.client import DarwinClient


class AVAudioSessionCategory(Enum):
    """https://developer.apple.com/documentation/avfaudio/AVAudioSessionCategory?language=objc"""

    PlayAndRecord = "AVAudioSessionCategoryPlayAndRecord"
    Ambient = "AVAudioSessionCategoryAmbient"
    MultiRoute = "AVAudioSessionCategoryMultiRoute"
    Playback = "AVAudioSessionCategoryPlayback"
    Record = "AVAudioSessionCategoryRecord"
    SoloAmbient = "AVAudioSessionCategorySoloAmbient"
    AudioProcessing = "AVAudioSessionCategoryAudioProcessing"


class AVAudioSessionMode(Enum):
    """https://developer.apple.com/documentation/avfaudio/avaudiosessionmode?language=objc"""

    Default = "AVAudioSessionModeDefault"
    GameChat = "AVAudioSessionModeGameChat"
    Measurement = "AVAudioSessionModeMeasurement"
    MoviePlayback = "AVAudioSessionModeMoviePlayback"
    SpokenAudio = "AVAudioSessionModeSpokenAudio"
    VideoChat = "AVAudioSessionModeVideoChat"
    VideoRecording = "AVAudioSessionModeVideoRecording"
    VoiceChat = "AVAudioSessionModeVoiceChat"
    VoicePrompt = "AVAudioSessionModeVoicePrompt"


class InterruptionPriority(Enum):
    Default = 0
    PhoneCall = 10
    EmergencyAlert = 20


class Recorder(Allocated):
    """
    Wrapper for AVAudioRecorder
    https://developer.apple.com/documentation/avfaudio/avaudiorecorder?language=objc
    """

    def __init__(self, client: "DarwinClient", session: "AudioSession", recorder: DarwinSymbol):
        super().__init__()
        self._client = client
        self._session = session
        self._recorder = recorder

    def _deallocate(self):
        self._recorder.objc_call("release")

    def record(self):
        self._session.set_category(AVAudioSessionCategory.PlayAndRecord)
        self._session.set_active(True)
        self._recorder.objc_call("record")
        if not self.recording:
            raise RpcFailedToRecordError()

    def pause(self):
        self._recorder.objc_call("pause")
        self._session.set_active(False)

    def stop(self):
        self._recorder.objc_call("stop")
        self._session.set_active(False)

    def delete_recording(self):
        if not self._recorder.objc_call("deleteRecording"):
            raise BadReturnValueError("deleteRecording failed")

    @property
    def recording(self) -> bool:
        return bool(self._recorder.objc_call("isRecording"))


class Player(Allocated):
    """
    Wrapper for AVAudioPlayer
    https://developer.apple.com/documentation/avfaudio/avaudioplayer?language=objc
    """

    def __init__(self, client: "DarwinClient", session: "AudioSession", player: DarwinSymbol):
        super().__init__()
        self._client = client
        self._session = session
        self._player = player

    def _deallocate(self):
        self._player.objc_call("release")

    def play(self):
        self._session.set_category(AVAudioSessionCategory.PlayAndRecord)
        self._session.set_active(True)
        self._player.objc_call("play")
        if not self.playing:
            raise RpcFailedToPlayError()

    def pause(self):
        self._player.objc_call("pause")

    def stop(self):
        self._player.objc_call("stop")
        self._session.set_active(False)

    def set_volume(self, value: float):
        self._player.objc_call("setVolume:", struct.pack("<f", value))

    @property
    def playing(self) -> bool:
        return bool(self._player.objc_call("isPlaying"))

    @property
    def loops(self) -> int:
        return self._player.objc_call("numberOfLoops")

    @loops.setter
    def loops(self, value: int):
        self._player.objc_call("setNumberOfLoops:", value)


class AudioSession:
    """
    wrapper for AVAudioSession
    https://developer.apple.com/documentation/avfaudio/avaudiosession?language=objc
    """

    def __init__(self, client: "DarwinClient"):
        self._client = client
        self._session = self._client.symbols.objc_getClass("AVAudioSession").objc_call("sharedInstance")

    def set_active(self, is_active: bool):
        self._session.objc_call("setActive:error:", is_active, 0)

    def set_mode(self, mode: AVAudioSessionMode):
        mode = self._client.symbols[mode.value][0]
        self._session.objc_call("setMode:error:", mode, 0)

    def set_category(
        self,
        category: AVAudioSessionCategory,
        mode: AVAudioSessionMode = AVAudioSessionMode.Default,
        route_sharing_policy: AVAudioSessionRouteSharingPolicy = AVAudioSessionRouteSharingPolicy.Default,
        options: AVAudioSessionCategoryOptions = AVAudioSessionCategoryOptions.DefaultToSpeaker,
    ) -> None:
        category = self._client.symbols[category.value][0]
        mode = self._client.symbols[mode.value][0]
        self._session.objc_call(
            "setCategory:mode:routeSharingPolicy:options:error:", category, mode, route_sharing_policy, options, 0
        )

    def set_interruption_priority(self, priority: InterruptionPriority):
        self._session.objc_call("setInterruptionPriority:error:", priority, 0)

    def override_output_audio_port(self, port: int):
        self._session.objc_call("overrideOutputAudioPort:error:", port, 0)

    @property
    def other_audio_playing(self) -> bool:
        return bool(self._session.objc_call("isOtherAudioPlaying"))

    @property
    def record_permission(self):
        return struct.pack("<I", self._session.objc_call("recordPermission"))[::-1].decode()

    @property
    def available_categories(self) -> list[str]:
        return self._session.objc_call("availableCategories").py()

    @property
    def available_modes(self) -> list[str]:
        return self._session.objc_call("availableModes").py()

    @property
    def is_active(self) -> bool:
        return bool(self._session.objc_call("isActive"))


class DarwinMedia:
    """Media utils"""

    def __init__(self, client: "DarwinClient"):
        """
        :param rpcclient.darwin.client.DarwinClient client:
        """
        self._client = client
        self._client.load_framework("AVFoundation")
        self.session = AudioSession(self._client)

    @path_to_str("filename")
    def get_recorder(self, filename: str) -> Recorder:
        url = self._client.symbols.objc_getClass("NSURL").objc_call("fileURLWithPath:", self._client.cf(filename))
        settings = self._client.cf({
            "AVEncoderQualityKey": 100,
            "AVEncoderBitRateKey": 16,
            "AVNumberOfChannelsKey": 1,
            "AVSampleRateKey": 8000.0,
        })
        AVAudioRecorder = self._client.symbols.objc_getClass("AVAudioRecorder")
        recorder = AVAudioRecorder.objc_call("alloc").objc_call("initWithURL:settings:error:", url, settings, 0)

        return Recorder(self._client, self.session, recorder)

    @path_to_str("filename")
    def get_player(self, filename: str) -> Player:
        NSURL = self._client.symbols.objc_getClass("NSURL")
        url = NSURL.objc_call("fileURLWithPath:", self._client.cf(filename))

        AVAudioPlayer = self._client.symbols.objc_getClass("AVAudioPlayer")
        player = AVAudioPlayer.objc_call("alloc").objc_call("initWithContentsOfURL:error:", url, 0)

        return Player(self._client, self.session, player)



from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from rpcclient.clients.darwin.client import DarwinClient


class Bluetooth:
    """bluetooth utils"""

    # bugfix: +[BluetoothManager.setSharedInstanceQueue:] cannot be called twice in the same process,
    # so we use this global to tell if it was already called
    _ENV_QUEUE_SET = "_rpc_server_bluetooth_manager_dispatch_queue_set"

    def __init__(self, client: "DarwinClient"):
        self._client = client
        self._client.load_framework("BluetoothManager")
        bluetooth_manager_class = client.symbols.objc_getClass("BluetoothManager")
        if not client.getenv(self._ENV_QUEUE_SET):
            bluetooth_manager_class.objc_call(
                "setSharedInstanceQueue:", self._client.symbols.dispatch_queue_create(0, 0)
            )
            client.setenv(self._ENV_QUEUE_SET, "1")
        self._bluetooth_manager = bluetooth_manager_class.objc_call("sharedInstance")

    def is_on(self):
        return self._bluetooth_manager.objc_call("enabled") == 1

    def turn_on(self):
        self._set(is_on=1)

    def turn_off(self):
        self._set(is_on=0)

    @property
    def address(self) -> Optional[str]:
        return self._bluetooth_manager.objc_call("localAddress").py()

    @property
    def connected(self) -> bool:
        return bool(self._bluetooth_manager.objc_call("connected"))

    @property
    def discoverable(self) -> bool:
        return bool(self._bluetooth_manager.objc_call("isDiscoverable"))

    @discoverable.setter
    def discoverable(self, value: bool):
        self._bluetooth_manager.objc_call("setDiscoverable:", value)

    def _set(self, is_on):
        self._bluetooth_manager.objc_call("setPowered:", is_on)
        self._bluetooth_manager.objc_call("setEnabled:", is_on)

    def __repr__(self):
        return f"<Bluetooth state:{'ON' if self.is_on() else 'OFF'}>"



import datetime

import pytest

pytestmark = pytest.mark.darwin


@pytest.mark.parametrize(
    "data",
    [
        None,
        True,
        False,
        datetime.datetime(1967, 5, 4, 3, 2, 1),
        "string",
        b"bytes",
        123,
        0.1,
        [0, 1, "abc"],
        {"key": "value"},
        [{"key": "value"}, [1, 2]],
    ],
)
def test_serialization(client, data):
    assert client.cf(data).py() == data



import logging
import socket
from typing import Optional

from rpcclient.exceptions import InvalidServerVersionMagicError, ServerResponseError
from rpcclient.protocol.messages import RpcMessageRegistry
from rpcclient.protocol.rpc_socket import RpcSocket
from rpcclient.protos.rpc_pb2 import ProtocolConstants, RpcMessage

logger = logging.getLogger(__name__)

BASIC_MESSAGES = RpcMessageRegistry(modules=["rpcclient.protos.rpc_api_pb2"])


class RpcBridge:
    def __init__(
        self,
        sock: RpcSocket,
        client_id: int,
        platform_name: str,
        arch: int,
        sysname: str,
        messages: Optional[RpcMessageRegistry] = None,
        owns_socket: bool = True,
    ) -> None:
        self.messages: RpcMessageRegistry = messages or BASIC_MESSAGES.clone()
        self.sock = sock
        self.client_id = client_id
        self.platform = platform_name
        self.arch = arch
        self.sysname = sysname
        self._owns_socket = owns_socket

    @classmethod
    def connect(cls, raw_sock: socket.socket, messages: Optional[RpcMessageRegistry] = None) -> "RpcBridge":
        sock = RpcSocket(raw_sock)
        handshake = sock.rpc_handshake_recv()
        if handshake.server_version != ProtocolConstants.SERVER_VERSION:
            raise InvalidServerVersionMagicError(
                f"got {handshake.magic:x} instead of {ProtocolConstants.SERVER_VERSION:x}"
            )
        return cls(
            sock, handshake.client_id, handshake.platform.lower(), handshake.arch, handshake.sysname.lower(), messages
        )

    def rpc_call(self, msg_id: int, **kwargs):
        """
        Resolve msg_id/reply class from request_msg's type, build RpcMessage, send and, parse reply.
        """
        req = self.messages.get(msg_id)(**kwargs)
        msg = RpcMessage(
            client_id=self.client_id,
            msg_id=msg_id,
            payload=req.SerializeToString(),
        )
        rep_msg = self.sock.rpc_msg_send_recv(msg)
        rep = self.messages.get(rep_msg.msg_id)()
        rep.ParseFromString(rep_msg.payload)
        if rep_msg.msg_id == ProtocolConstants.REP_ERROR:
            logger.error(f"Server error: {rep.message}")
            raise ServerResponseError(rep.message)
        return rep

    def close(self) -> None:
        if not self._owns_socket:
            raise RuntimeError("socket is owned by another client")
        self.sock.raw_socket.close()

    def clone(self) -> "RpcBridge":
        return RpcBridge(
            self.sock, self.client_id, self.platform, self.arch, self.sysname, self.messages.clone(), owns_socket=False
        )



def test_valid_gethostbyname(client):
    """
    :param rpcclient.client.Client client:
    """
    assert client.network.gethostbyname("google.com") is not None


def test_invalid_gethostbyname(client):
    """
    :param rpcclient.client.Client client:
    """
    assert client.network.gethostbyname("google.com1") is None



import logging
from typing import TYPE_CHECKING, Any, Optional
from uuid import UUID

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from rpcclient.clients.ios.client import IosClient


class Call:
    """Represents a single ongoing call."""

    def __init__(self, client: "IosClient", controller: Any, call: Any) -> None:
        self._client = client
        self._controller = controller
        self._call = call

    def disconnect(self) -> None:
        """Disconnect the current call."""
        self._send_action("CXEndCallAction")

    def answer(self) -> None:
        """Answer the current call."""
        self._send_action("CXAnswerCallAction")

    def _send_action(self, action_name: str) -> None:
        action_class = self._client.symbols.objc_getClass(action_name)
        action = action_class.objc_call("alloc").objc_call("initWithCallUUID:", self._uuid)
        self._controller.objc_call("requestTransactionWithAction:completion:", action, self._client.get_dummy_block())

    @property
    def _uuid(self) -> UUID:
        return self._call.objc_call("UUID")


class Telephony:
    """
    Telephony utilities backed by CallKit and CoreTelephony.

    Accessing real telephony requires the entitlement "application-identifier"
    to be set to "com.apple.coretelephony".
    """

    def __init__(self, client: "IosClient") -> None:
        self._client = client
        self._client.load_framework("CallKit")
        self.cx_call_controller = self._client.symbols.objc_getClass("CXCallController").objc_call("new")
        self.cx_call_observer = self.cx_call_controller.objc_call("callObserver")
        self.ct_message_center = self._client.symbols.objc_getClass("CTMessageCenter").objc_call("sharedMessageCenter")

    def dial(self, number: str) -> None:
        """
        Start a call to a number.

        Use `current_call` to access the created call.
        """
        self._client.symbols.CTCallDial(self._client.cf(number))

    def send_sms(self, to_address: str, text: str, smsc: str = "1111") -> None:
        """
        Send an SMS message.

        :param to_address: Destination phone address.
        :param text: Message text.
        :param smsc: Originator short message service center address.
        """
        self.ct_message_center.objc_call(
            "sendSMSWithText:serviceCenter:toAddress:",
            self._client.cf(text),
            self._client.cf(smsc),
            self._client.cf(to_address),
        )

    @property
    def current_call(self) -> Optional[Call]:
        """
        Return an object representing the current active call, if any.
        """
        calls = self.cx_call_observer.objc_call("calls")
        call_count = calls.objc_call("count")

        call_list = []
        for i in range(call_count):
            call_list.append(calls.objc_call("objectAtIndex:", i))

        for call_id in range(call_count):
            call = call_list[call_id]
            if call.objc_call("hasEnded"):
                continue
            return Call(self._client, self.cx_call_controller, call)

        return None



from typing import TYPE_CHECKING

from rpcclient.clients.darwin.common import CfSerializable

if TYPE_CHECKING:
    from rpcclient.clients.ios.client import IosClient


class MobileGestalt:
    """Thin wrapper around MobileGestalt MGCopyAnswer/MGSetAnswer keys."""

    def __init__(self, client: "IosClient") -> None:
        self._client = client

    # Identifying Information

    @property
    def DiskUsage(self) -> CfSerializable:
        return self.get_answer("DiskUsage")

    @property
    def ModelNumber(self) -> CfSerializable:
        return self.get_answer("ModelNumber")

    @property
    def SIMTrayStatus(self) -> CfSerializable:
        return self.get_answer("SIMTrayStatus")

    @property
    def SerialNumber(self) -> CfSerializable:
        return self.get_answer("SerialNumber")

    @property
    def MLBSerialNumber(self) -> CfSerializable:
        return self.get_answer("MLBSerialNumber")

    @property
    def UniqueDeviceID(self) -> CfSerializable:
        return self.get_answer("UniqueDeviceID")

    @property
    def UniqueDeviceIDData(self) -> CfSerializable:
        return self.get_answer("UniqueDeviceIDData")

    @property
    def UniqueChipID(self) -> CfSerializable:
        return self.get_answer("UniqueChipID")

    @property
    def InverseDeviceID(self) -> CfSerializable:
        return self.get_answer("InverseDeviceID")

    @property
    def DiagData(self) -> CfSerializable:
        return self.get_answer("DiagData")

    @property
    def DieId(self) -> CfSerializable:
        return self.get_answer("DieId")

    @property
    def CPUArchitecture(self) -> CfSerializable:
        return self.get_answer("CPUArchitecture")

    @property
    def PartitionType(self) -> CfSerializable:
        return self.get_answer("PartitionType")

    @property
    def UserAssignedDeviceName(self) -> CfSerializable:
        return self.get_answer("UserAssignedDeviceName")

    # Bluetooth Information

    @property
    def BluetoothAddress(self) -> CfSerializable:
        return self.get_answer("BluetoothAddress")

    # Battery Information

    @property
    def RequiredBatteryLevelForSoftwareUpdate(self) -> CfSerializable:
        return self.get_answer("RequiredBatteryLevelForSoftwareUpdate")

    @property
    def BatteryIsFullyCharged(self) -> CfSerializable:
        return self.get_answer("BatteryIsFullyCharged")

    @property
    def BatteryIsCharging(self) -> CfSerializable:
        return self.get_answer("BatteryIsCharging")

    @property
    def BatteryCurrentCapacity(self) -> CfSerializable:
        return self.get_answer("BatteryCurrentCapacity")

    @property
    def ExternalPowerSourceConnected(self) -> CfSerializable:
        return self.get_answer("ExternalPowerSourceConnected")

    # Baseband Information

    @property
    def BasebandSerialNumber(self) -> CfSerializable:
        return self.get_answer("BasebandSerialNumber")

    @property
    def BasebandCertId(self) -> CfSerializable:
        return self.get_answer("BasebandCertId")

    @property
    def BasebandChipId(self) -> CfSerializable:
        return self.get_answer("BasebandChipId")

    @property
    def BasebandFirmwareManifestData(self) -> CfSerializable:
        return self.get_answer("BasebandFirmwareManifestData")

    @property
    def BasebandFirmwareVersion(self) -> CfSerializable:
        return self.get_answer("BasebandFirmwareVersion")

    @property
    def BasebandKeyHashInformation(self) -> CfSerializable:
        return self.get_answer("BasebandKeyHashInformation")

    # Telephony Information

    @property
    def CarrierBundleInfoArray(self) -> CfSerializable:
        return self.get_answer("CarrierBundleInfoArray")

    @property
    def CarrierInstallCapability(self) -> CfSerializable:
        return self.get_answer("CarrierInstallCapability")

    @property
    def InternationalMobileEquipmentIdentity(self) -> CfSerializable:
        return self.get_answer("InternationalMobileEquipmentIdentity")

    @property
    def MobileSubscriberCountryCode(self) -> CfSerializable:
        return self.get_answer("MobileSubscriberCountryCode")

    @property
    def MobileSubscriberNetworkCode(self) -> CfSerializable:
        return self.get_answer("MobileSubscriberNetworkCode")

    # Device Information

    @property
    def ChipID(self) -> CfSerializable:
        return self.get_answer("ChipID")

    @property
    def ComputerName(self) -> CfSerializable:
        return self.get_answer("ComputerName")

    @property
    def DeviceVariant(self) -> CfSerializable:
        return self.get_answer("DeviceVariant")

    @property
    def HWModelStr(self) -> CfSerializable:
        return self.get_answer("HWModelStr")

    @property
    def BoardId(self) -> CfSerializable:
        return self.get_answer("BoardId")

    @property
    def HardwarePlatform(self) -> CfSerializable:
        return self.get_answer("HardwarePlatform")

    @property
    def DeviceName(self) -> CfSerializable:
        return self.get_answer("DeviceName")

    @property
    def DeviceColor(self) -> CfSerializable:
        return self.get_answer("DeviceColor")

    @property
    def DeviceClassNumber(self) -> CfSerializable:
        return self.get_answer("DeviceClassNumber")

    @property
    def DeviceClass(self) -> CfSerializable:
        return self.get_answer("DeviceClass")

    @property
    def BuildVersion(self) -> CfSerializable:
        return self.get_answer("BuildVersion")

    @property
    def ProductName(self) -> CfSerializable:
        return self.get_answer("ProductName")

    @property
    def ProductType(self) -> CfSerializable:
        return self.get_answer("ProductType")

    @property
    def ProductVersion(self) -> CfSerializable:
        return self.get_answer("ProductVersion")

    @property
    def FirmwareNonce(self) -> CfSerializable:
        return self.get_answer("FirmwareNonce")

    @property
    def FirmwareVersion(self) -> CfSerializable:
        return self.get_answer("FirmwareVersion")

    @property
    def FirmwarePreflightInfo(self) -> CfSerializable:
        return self.get_answer("FirmwarePreflightInfo")

    @property
    def IntegratedCircuitCardIdentifier(self) -> CfSerializable:
        return self.get_answer("IntegratedCircuitCardIdentifier")

    @property
    def AirplaneMode(self) -> bool:
        return self.get_answer("AirplaneMode")

    @property
    def AllowYouTube(self) -> CfSerializable:
        return self.get_answer("AllowYouTube")

    @property
    def AllowYouTubePlugin(self) -> CfSerializable:
        return self.get_answer("AllowYouTubePlugin")

    @property
    def MinimumSupportediTunesVersion(self) -> CfSerializable:
        return self.get_answer("MinimumSupportediTunesVersion")

    @property
    def ProximitySensorCalibration(self) -> CfSerializable:
        return self.get_answer("ProximitySensorCalibration")

    @property
    def RegionCode(self) -> CfSerializable:
        return self.get_answer("RegionCode")

    @property
    def RegionInfo(self) -> CfSerializable:
        return self.get_answer("RegionInfo")

    @property
    def RegulatoryIdentifiers(self) -> CfSerializable:
        return self.get_answer("RegulatoryIdentifiers")

    @property
    def SBAllowSensitiveUI(self) -> CfSerializable:
        return self.get_answer("SBAllowSensitiveUI")

    @property
    def SBCanForceDebuggingInfo(self) -> CfSerializable:
        return self.get_answer("SBCanForceDebuggingInfo")

    @property
    def SDIOManufacturerTuple(self) -> CfSerializable:
        return self.get_answer("SDIOManufacturerTuple")

    @property
    def SDIOProductInfo(self) -> CfSerializable:
        return self.get_answer("SDIOProductInfo")

    @property
    def ShouldHactivate(self) -> CfSerializable:
        return self.get_answer("ShouldHactivate")

    @property
    def SigningFuse(self) -> CfSerializable:
        return self.get_answer("SigningFuse")

    @property
    def SoftwareBehavior(self) -> CfSerializable:
        return self.get_answer("SoftwareBehavior")

    @property
    def SoftwareBundleVersion(self) -> CfSerializable:
        return self.get_answer("SoftwareBundleVersion")

    @property
    def SupportedDeviceFamilies(self) -> CfSerializable:
        return self.get_answer("SupportedDeviceFamilies")

    @property
    def SupportedKeyboards(self) -> CfSerializable:
        return self.get_answer("SupportedKeyboards")

    @property
    def TotalSystemAvailable(self) -> CfSerializable:
        return self.get_answer("TotalSystemAvailable")

    # Capability Information

    @property
    def AllDeviceCapabilities(self) -> CfSerializable:
        return self.get_answer("AllDeviceCapabilities")

    @property
    def AppleInternalInstallCapability(self) -> CfSerializable:
        return self.get_answer("AppleInternalInstallCapability")

    @property
    def ExternalChargeCapability(self) -> CfSerializable:
        return self.get_answer("ExternalChargeCapability")

    @property
    def ForwardCameraCapability(self) -> CfSerializable:
        return self.get_answer("ForwardCameraCapability")

    @property
    def PanoramaCameraCapability(self) -> CfSerializable:
        return self.get_answer("PanoramaCameraCapability")

    @property
    def RearCameraCapability(self) -> CfSerializable:
        return self.get_answer("RearCameraCapability")

    @property
    def HasAllFeaturesCapability(self) -> CfSerializable:
        return self.get_answer("HasAllFeaturesCapability")

    @property
    def HasBaseband(self) -> CfSerializable:
        return self.get_answer("HasBaseband")

    @property
    def HasInternalSettingsBundle(self) -> CfSerializable:
        return self.get_answer("HasInternalSettingsBundle")

    @property
    def HasSpringBoard(self) -> CfSerializable:
        return self.get_answer("HasSpringBoard")

    @property
    def InternalBuild(self) -> CfSerializable:
        return self.get_answer("InternalBuild")

    @property
    def IsSimulator(self) -> CfSerializable:
        return self.get_answer("IsSimulator")

    @property
    def IsThereEnoughBatteryLevelForSoftwareUpdate(self) -> CfSerializable:
        return self.get_answer("IsThereEnoughBatteryLevelForSoftwareUpdate")

    @property
    def IsUIBuild(self) -> CfSerializable:
        return self.get_answer("IsUIBuild")

    @property
    def PasswordConfigured(self) -> CfSerializable:
        return self.get_answer("PasswordConfigured")

    @property
    def PasswordProtected(self) -> CfSerializable:
        return self.get_answer("PasswordProtected")

    # Regional Behaviour

    @property
    def RegionalBehaviorAll(self) -> CfSerializable:
        return self.get_answer("RegionalBehaviorAll")

    @property
    def RegionalBehaviorChinaBrick(self) -> CfSerializable:
        return self.get_answer("RegionalBehaviorChinaBrick")

    @property
    def RegionalBehaviorEUVolumeLimit(self) -> CfSerializable:
        return self.get_answer("RegionalBehaviorEUVolumeLimit")

    @property
    def RegionalBehaviorGB18030(self) -> CfSerializable:
        return self.get_answer("RegionalBehaviorGB18030")

    @property
    def RegionalBehaviorGoogleMail(self) -> CfSerializable:
        return self.get_answer("RegionalBehaviorGoogleMail")

    @property
    def RegionalBehaviorNTSC(self) -> CfSerializable:
        return self.get_answer("RegionalBehaviorNTSC")

    @property
    def RegionalBehaviorNoPasscodeLocationTiles(self) -> CfSerializable:
        return self.get_answer("RegionalBehaviorNoPasscodeLocationTiles")

    @property
    def RegionalBehaviorNoVOIP(self) -> CfSerializable:
        return self.get_answer("RegionalBehaviorNoVOIP")

    @property
    def RegionalBehaviorNoWiFi(self) -> CfSerializable:
        return self.get_answer("RegionalBehaviorNoWiFi")

    @property
    def RegionalBehaviorShutterClick(self) -> CfSerializable:
        return self.get_answer("RegionalBehaviorShutterClick")

    @property
    def RegionalBehaviorVolumeLimit(self) -> CfSerializable:
        return self.get_answer("RegionalBehaviorVolumeLimit")

    # Wireless Information

    @property
    def ActiveWirelessTechnology(self) -> CfSerializable:
        return self.get_answer("ActiveWirelessTechnology")

    @property
    def WifiAddress(self) -> CfSerializable:
        return self.get_answer("WifiAddress")

    @property
    def WifiAddressData(self) -> CfSerializable:
        return self.get_answer("WifiAddressData")

    @property
    def WifiVendor(self) -> CfSerializable:
        return self.get_answer("WifiVendor")

    # FaceTime Information

    @property
    def FaceTimeBitRate2G(self) -> CfSerializable:
        return self.get_answer("FaceTimeBitRate2G")

    @property
    def FaceTimeBitRate3G(self) -> CfSerializable:
        return self.get_answer("FaceTimeBitRate3G")

    @property
    def FaceTimeBitRateLTE(self) -> CfSerializable:
        return self.get_answer("FaceTimeBitRateLTE")

    @property
    def FaceTimeBitRateWiFi(self) -> CfSerializable:
        return self.get_answer("FaceTimeBitRateWiFi")

    @property
    def FaceTimeDecodings(self) -> CfSerializable:
        return self.get_answer("FaceTimeDecodings")

    @property
    def FaceTimeEncodings(self) -> CfSerializable:
        return self.get_answer("FaceTimeEncodings")

    @property
    def FaceTimePreferredDecoding(self) -> CfSerializable:
        return self.get_answer("FaceTimePreferredDecoding")

    @property
    def FaceTimePreferredEncoding(self) -> CfSerializable:
        return self.get_answer("FaceTimePreferredEncoding")

    # More Device Capabilities

    @property
    def DeviceSupportsFaceTime(self) -> CfSerializable:
        return self.get_answer("DeviceSupportsFaceTime")

    @property
    def DeviceSupportsTethering(self) -> CfSerializable:
        return self.get_answer("DeviceSupportsTethering")

    @property
    def DeviceSupportsSimplisticRoadMesh(self) -> CfSerializable:
        return self.get_answer("DeviceSupportsSimplisticRoadMesh")

    @property
    def DeviceSupportsNavigation(self) -> CfSerializable:
        return self.get_answer("DeviceSupportsNavigation")

    @property
    def DeviceSupportsLineIn(self) -> CfSerializable:
        return self.get_answer("DeviceSupportsLineIn")

    @property
    def DeviceSupports9Pin(self) -> CfSerializable:
        return self.get_answer("DeviceSupports9Pin")

    @property
    def DeviceSupports720p(self) -> CfSerializable:
        return self.get_answer("DeviceSupports720p")

    @property
    def DeviceSupports4G(self) -> CfSerializable:
        return self.get_answer("DeviceSupports4G")

    @property
    def DeviceSupports3DMaps(self) -> CfSerializable:
        return self.get_answer("DeviceSupports3DMaps")

    @property
    def DeviceSupports3DImagery(self) -> CfSerializable:
        return self.get_answer("DeviceSupports3DImagery")

    @property
    def DeviceSupports1080p(self) -> CfSerializable:
        return self.get_answer("DeviceSupports1080p")

    def get_answer(self, key: str) -> CfSerializable:
        """Return the MGCopyAnswer value for a given key."""
        return self._client.symbols.MGCopyAnswer(self._client.cf(key)).py()

    def set_answer(self, key: str, value: CfSerializable) -> CfSerializable:
        """Set the MGSetAnswer value for a given key."""
        return self._client.symbols.MGSetAnswer(self._client.cf(key), self._client.cf(value))



import logging
from typing import TYPE_CHECKING

from rpcclient.core.subsystems.lief import Symbol
from rpcclient.exceptions import BadReturnValueError, RpcPermissionError

if TYPE_CHECKING:
    from rpcclient.clients.darwin.client import DarwinClient

logger = logging.getLogger(__name__)


class Keychain:
    """keychain utils"""

    def __init__(self, client: "DarwinClient"):
        self._client = client

    def add_internet_password(self, account: str, server: str, password: str):
        attributes = self._client.symbols.objc_getClass("NSMutableDictionary").objc_call("new")
        attributes.objc_call(
            "setObject:forKey:", self._client.symbols.kSecClassInternetPassword[0], self._client.symbols.kSecClass[0]
        )
        attributes.objc_call("setObject:forKey:", self._client.cf(account), self._client.symbols.kSecAttrAccount[0])
        attributes.objc_call("setObject:forKey:", self._client.cf(server), self._client.symbols.kSecAttrServer[0])
        attributes.objc_call("setObject:forKey:", self._client.cf(password), self._client.symbols.kSecValueData[0])
        err = self._client.symbols.SecItemAdd(attributes, 0).c_int32
        if err != 0:
            raise BadReturnValueError(f"SecItemAdd() returned: {err}")

    def query_apple_share_passwords(self) -> list[dict]:
        return self._query(self._client.symbols.kSecClassAppleSharePassword)

    def query_internet_passwords(self) -> list[dict]:
        return self._query(self._client.symbols.kSecClassInternetPassword)

    def query_generic_passwords(self) -> list[dict]:
        return self._query(self._client.symbols.kSecClassGenericPassword)

    def query_identities(self) -> list[dict]:
        return self._query(self._client.symbols.kSecClassIdentity)

    def query_certificates(self) -> list[dict]:
        return self._query(self._client.symbols.kSecClassCertificate)

    def query_keys(self) -> list[dict]:
        return self._query(self._client.symbols.kSecClassKey)

    def _query(self, class_type: Symbol) -> list[dict]:
        with self._client.safe_malloc(8) as p_result:
            p_result[0] = 0

            query = self._client.symbols.objc_getClass("NSMutableDictionary").objc_call("new")
            query.objc_call("setObject:forKey:", class_type[0], self._client.symbols.kSecClass[0])
            query.objc_call(
                "setObject:forKey:", self._client.symbols.kSecMatchLimitAll[0], self._client.symbols.kSecMatchLimit[0]
            )
            query.objc_call(
                "setObject:forKey:",
                self._client.symbols.kCFBooleanTrue[0],
                self._client.symbols.kSecReturnAttributes[0],
            )
            query.objc_call(
                "setObject:forKey:", self._client.symbols.kCFBooleanTrue[0], self._client.symbols.kSecReturnRef[0]
            )
            query.objc_call(
                "setObject:forKey:", self._client.symbols.kCFBooleanTrue[0], self._client.symbols.kSecReturnData[0]
            )

            err = self._client.symbols.SecItemCopyMatching(query, p_result).c_int32
            if err != 0:
                raise BadReturnValueError(f"SecItemCopyMatching() returned: {err}")

            result = p_result[0]

            if result == 0:
                raise RpcPermissionError()

            # results contain a reference which isn't plist-serializable
            keys_to_remove = [self._client.cf("v_Ref"), self._client.cf("accc")]
            for i in range(result.objc_call("count")):
                for removal_key in keys_to_remove:
                    result.objc_call("objectAtIndex:", i).objc_call("removeObjectForKey:", removal_key)
            return result.py()



import dataclasses
import logging
import sys
from typing import Any, Optional, Union

import click
import IPython
from IPython import get_ipython
from traitlets.config import Config

from rpcclient.client_manager import ClientManager, ClientType
from rpcclient.console.extensions.keybindings import get_keybindings
from rpcclient.exceptions import NoSuchClientError
from rpcclient.registry import Registry, RegistryEvent
from rpcclient.utils import prompt_selection

logger = logging.getLogger(__name__)


def disable_loggers() -> None:
    """Disable noisy third-party loggers for a cleaner shell."""
    logging.getLogger("asyncio").disabled = True
    logging.getLogger("parso").disabled = True
    logging.getLogger("parso.cache").disabled = True
    logging.getLogger("parso.python.diff").disabled = True
    logging.getLogger("blib2to3.pgen2.driver").disabled = True
    logging.getLogger("humanfriendly.prompts").disabled = True
    logging.getLogger("urllib3.connectionpool").disabled = True
    logging.getLogger("rpcclient.event_dispatcher").disabled = True
    logging.getLogger("rpcclient.client_manager").disabled = True


@dataclasses.dataclass
class HelpSnippet:
    """Single help line for console shortcuts."""

    key: str
    description: str

    def __str__(self) -> str:
        return click.style(self.key.ljust(8), bold=True, fg="magenta") + click.style(self.description, bold=True)


@dataclasses.dataclass
class ConsoleContext:
    """Per-client console state (active client and its user namespace)."""

    p: ClientType
    user_ns: dict[str, Any] = dataclasses.field(default_factory=dict)

    @property
    def id(self) -> int:
        """Return the client ID of the client."""
        return self.p.id

    def __repr__(self) -> str:
        """Debug-friendly representation of the client."""
        return repr(self.p)


def echo_info(msg: str) -> None:
    """Print an informational message in cyan."""
    click.secho(msg, fg="cyan")


def echo_warn(msg: str) -> None:
    """Print a warning message in yellow."""
    click.secho(f"⚠ {msg}", fg="yellow")


def echo_error(msg: str) -> None:
    """Print an error message in red."""
    click.secho(f"✖ {msg}", fg="red", err=True)


def echo_success(msg: str) -> None:
    """Print a success message in green."""
    click.secho(f"✔ {msg}", fg="green")


def echo_heading(msg: str) -> None:
    """Print a highlighted heading."""
    click.secho(msg, fg="magenta", bold=True)


class Console:
    """Interactive IPython console orchestrating multiple RPC clients."""

    def __init__(self, mgr: ClientManager) -> None:
        """Initialize the console with a client manager and event wiring."""
        self.mgr = mgr
        self._contexts = Registry()
        self._current: Union[int, None] = None
        self._previous: Union[int, None] = None
        self._ipython = None
        self._baseline_keys = None
        self._auto_switch_on_create = True
        self._setup_done = False

        self.mgr.notifier.register(RegistryEvent.REGISTERED, self._on_registered)
        self.mgr.notifier.register(RegistryEvent.UNREGISTERED, self._on_unregistered)
        self.mgr.notifier.register(RegistryEvent.CLEARED, self._on_clear)

    @property
    def contexts(self) -> dict[int, ConsoleContext]:
        return dict(self._contexts.items())

    def init_contexts(self) -> None:
        """Seed console contexts from current live clients."""
        for cid, client in self.mgr.clients.items():
            if cid in self._contexts:
                continue
            self._contexts.register(cid, ConsoleContext(client))

    def setup_shell(self, switch_cid: Union[int, None] = None) -> None:
        """One-time shell setup: quiet logs, capture baseline, and optionally switch."""
        if self._setup_done:
            return
        disable_loggers()
        self._ipython = get_ipython()
        self._baseline_keys = set(self._ipython.user_ns.keys())
        self.init_contexts()
        if switch_cid is not None:
            self.switch(switch_cid)
        self._setup_done = True

    def interactive(
        self,
        additional_namespace: Optional[dict] = None,
        switch_cid: Optional[int] = None,
        startup_files: Optional[tuple[str]] = None,
    ) -> None:
        """Launch the IPython shell with custom config and extensions."""
        sys.argv = ["a"]
        ipython_config = Config()
        ipython_config.IPCompleter.use_jedi = False
        ipython_config.InteractiveShellApp.exec_lines = [f"""console.setup_shell({switch_cid})"""]
        ipython_config.TerminalInteractiveShell.autoformatter = None
        ipython_config.BaseIPythonApplication.profile = "rpcclient"
        ipython_config.TerminalInteractiveShell.prompts_class = "rpcclient.console.prompt.RpcPrompt"
        ipython_config.InteractiveShellApp.extensions = [
            "rpcclient.console.extensions.events",
            "rpcclient.console.extensions.keybindings",
        ]
        if (startup_files is not None) and (len(startup_files) > 0):
            ipython_config.InteractiveShellApp.exec_files = startup_files

        namespace = {"console": self, "mgr": self.mgr}
        if additional_namespace is not None:
            namespace.update(additional_namespace)

        click.secho("RpcClient has been successfully loaded! 😎", bold=True)
        click.secho("Usage:", bold=True)
        self.show_help()
        click.echo(click.style("Have a nice flight ✈️! Starting an IPython shell...", bold=True))

        if (startup_files is not None) and (len(startup_files) > 0):
            click.secho(f"Detected Startup files to be run: {startup_files}", bold=True)

        IPython.start_ipython(config=ipython_config, user_ns=namespace)
        self.mgr.clear()

    def switch(self, cid: Union[int, None] = None) -> None:
        """Switch the active console context by client ID (or interactively pick one)."""
        if not self._contexts.items():
            raise NoSuchClientError("No clients available")

        if cid is None:
            choices = [ctx for _, ctx in self._contexts.items()]
            ctx = prompt_selection(choices, "Select a client client ID")
        else:
            ctx = self._contexts.get(cid)

        if ctx is None:
            client = self.mgr.get(cid)
            if client is None:
                echo_error(f"No such client: {cid}")
                return
            ctx = ConsoleContext(client)
            self._contexts.register(client.id, ctx)

        self._switch(ctx)
        echo_success(f"Switched to client client ID: {ctx.id}")

    def _switch(self, ctx: ConsoleContext) -> None:
        """Internal: swap `p` and per-context variables in the shell namespace."""
        ns = self._ipython.user_ns
        base = self._baseline_keys

        if self._current is not None and self._current in self._contexts:
            cur = self._contexts.get(self._current)
            cur.user_ns = {k: v for k, v in ns.items() if k not in base and k not in ("p", "symbols")}

        for k in list(ns.keys()):
            if k not in base:
                ns.pop(k, None)

        ns.update(ctx.user_ns)
        ns["p"] = ctx.p

        self._previous = self._current
        self._current = ctx.id

    def _reset_current(self, msg: str = "Current client has been removed") -> None:
        """Clear current client and console binding, then warn."""
        self._current = None
        self._ipython.user_ns["p"] = None
        echo_warn(msg)

    # ---------------------------------------------------------------------------
    # Client manager — event handlers
    # ---------------------------------------------------------------------------

    def _on_registered(self, cid, client: ClientType, internal: bool) -> None:
        """Event: auto-switch to a newly created client if enabled."""
        if not (self._auto_switch_on_create and self._ipython):
            return
        ctx = ConsoleContext(client)
        self._contexts.register(cid, ctx)
        if internal:
            # Prevent auto-switch on internal events as they are not client-facing
            return
        self._switch(ctx)
        echo_info(f"Auto-switched to new client client ID: {client.id}")

    def _on_unregistered(self, cid: int) -> None:
        """Event: clean up context for a removed client."""
        if cid not in self._contexts:
            return
        self._contexts.unregister(cid)
        if self._current != cid:
            return
        self._reset_current()

    def _on_clear(self) -> None:
        """Event: clean up all contexts."""
        self._contexts.clear()
        self._reset_current()

    # ---------------------------------------------------------------------------
    # Keybinding callbacks
    # ---------------------------------------------------------------------------

    def show_help(self, *_: Any) -> None:
        """Print console help and keybindings."""
        help_snippets = [
            HelpSnippet(key="mgr", description="Client manager: create | get | remove | clients | clear"),
            HelpSnippet(key="console", description="Console controller: switch"),
            HelpSnippet(key="p", description="Active client (e.g., p.info(), p.pid)"),
        ]
        for keybinding in get_keybindings(self):
            help_snippets.append(HelpSnippet(key=keybinding.key.upper(), description=keybinding.description))

        for help_snippet in help_snippets:
            click.echo(help_snippet)

    def toggle_auto_switch(self, *_: Any) -> None:
        """Toggle the auto-switch behavior on client creation."""
        self._auto_switch_on_create = not self._auto_switch_on_create
        status = "enabled" if self._auto_switch_on_create else "disabled"
        echo_info(f"Auto-switch on client creation: {status}")

    def previous_context(self, event: Any) -> None:
        """Switch back to the previous active context."""
        if self._previous is None or self._previous not in self._contexts:
            echo_error("No previous context available.")
            return
        self.switch(self._previous)

        # The prompt doesn't change on its own (no In/Out updates) - commenting it out and executing ("Enter" simulation) to trigger a prompt change
        Console._comment_input_and_execute(event)

    def show_contexts(self, *_: Any) -> None:
        """List active contexts in a human-readable form."""
        click.echo("\n".join([str(ctx.p) for _, ctx in self.contexts.items()]))

    @staticmethod
    def _comment_input_and_execute(event: Any):
        """
        Comments out the input, single-line or multi-line, and executes it (like pressing "Enter")
        """
        buffer = event.current_buffer
        text = buffer.text

        # Prepend '#' to every non-comment line in the buffer
        commented_text = "\n".join([f"# {line}" for line in text.splitlines() if not line.startswith("#")])

        # Update the buffer with the commented text
        buffer.text = commented_text

        # Move the cursor to the end to ensure smooth execution
        buffer.cursor_position = len(commented_text)

        # Finally, validate and execute
        buffer.validate_and_handle()



from rpcclient.clients.darwin.client import DarwinClient
from rpcclient.clients.darwin.subsystems.reports import Reports
from rpcclient.clients.macos.subsystems.apple_script import AppleScript
from rpcclient.core.subsystems.decorator import subsystem

CRASH_REPORTS_DIR = "Library/Logs/DiagnosticReports"


class MacosClient(DarwinClient):
    @subsystem
    def reports(self) -> Reports:
        return Reports(self, CRASH_REPORTS_DIR)

    @subsystem
    def apple_script(self) -> AppleScript:
        return AppleScript(self)

    @property
    def roots(self) -> list[str]:
        """get a list of all accessible darwin roots when used for lookup of files/preferences/..."""

        result = super().roots
        for username in self.fs.scandir("/Users"):
            if not username.is_dir() or not self.fs.accessible(username.path):
                continue
            result.append(username.path)
        return result
